# This reimplementation was based on the code found in http://cfgsim.cs.arizona.edu/. The code found there is an
# implementation based on the following paper:
# "Large-Scale Malware Indexing Using Function-Call Graphs"
# by Hu et al. and published at CCS 2009.
#
# This reimplementation differs in how it handles the cost of relabeling nodes. The original paper
# uses instruction information to decide a score. We dont implement that here, and instead we add the extra
# constraint of illegal edits to the graph (swapping a start or exit with a non-start or non-exit).
#

import math
import sys

import networkx as nx

from cfgutils.matrix.munkres import Munkres
from cfgutils.similarity.ged import INVALID_CHOICE_PENALTY


class GraphCache:
    def __init__(self, graph: nx.DiGraph):
        self._graph = graph
        self.nodes = list(graph.nodes)
        self.node_count = len(self.nodes)
        self.node_to_index = {node: i for i, node in enumerate(graph.nodes)}

        self._parents = {node: list(graph.predecessors(node)) for node in graph.nodes}
        self.parent_count = {node: len(parents) for node, parents in self._parents.items()}

        self._children = {node: list(graph.successors(node)) for node in graph.nodes}
        self.child_count = {node: len(children) for node, children in self._children.items()}

    def get_parent(self, node, i):
        return self._parents[node][i]

    def parent_node_idx(self, node, parent_i):
        return self.node_to_index[self.get_parent(node, parent_i)]

    def get_child(self, node, i):
        return self._children[node][i]

    def child_node_idx(self, node, child_i):
        return self.node_to_index[self.get_child(node, child_i)]


class CFGSimNM:
    def __init__(self, g1, g2, print_steps=False, normalize=False):
        self.__print_steps = print_steps
        self._normalize = normalize

        self.__eps = 0.0001
        self.__inf = float('inf')

        self._g1_graph = g1
        self._g2_graph = g2
        self._g1 = GraphCache(self._g1_graph)
        self._g2 = GraphCache(self._g2_graph)

    def relabel_cost(self, n1_idx, n2_idx):
        """
        This was previously contrastByInstr, which became contrast_by_instr, which became this.
        In the paper this algorithm is from the authors used it to decrease the cost of some operations if two
        nodes matched. If this function is not changed, the default cost is 1.

        In the case of CFGUtils, we consider swapping a start or exit with a non-start or non-exit to be an
        invalid choice. That is reflected in the cost returned by this function.
        """
        n1 = self._g1.nodes[n1_idx]
        n2 = self._g2.nodes[n2_idx]

        cost = 1
        if n1.is_entrypoint and not n2.is_entrypoint:
            cost = INVALID_CHOICE_PENALTY
        elif n1.is_exitpoint and not n2.is_exitpoint:
            cost = INVALID_CHOICE_PENALTY
        elif n2.is_entrypoint and not n1.is_entrypoint:
            cost = INVALID_CHOICE_PENALTY
        elif n2.is_exitpoint and not n1.is_exitpoint:
            cost = INVALID_CHOICE_PENALTY

        return cost

    def sim(self):
        n = self._g1.node_count
        m = self._g2.node_count

        sim = [None] * n
        for i in range(n):
            sim[i] = [1] * m
            for j in range(m):
                sim[i][j] = self.relabel_cost(i, j)

        if self.__print_steps:
            print('Initial cost matrix:')
            for i in range(n):
                for j in range(m):
                    sys.stdout.write(str(round(sim[i][j], 2)) + '\t')
                print('')

        finished = False    
        while not finished:
            # Since the calculation of the new sim matrix is based on the old one, we need to make
            # a copy of the old one first.
            old_sim = [None] * n
            for i in range(n):
                old_sim[i] = list(sim[i])
    
            precise = True
    
            for i in range(n):
                for j in range(m):
                    if self.__print_steps:
                        print('Current cost matrix:')
                        for k in range(n):
                            for l in range(m):
                                sys.stdout.write(str(round(sim[k][l], 2)))
                                if l < m - 1:
                                    sys.stdout.write(' & ')
                            print(' \\\\')
                        print('')
            
                    n1 = self._g1.nodes[i]
                    n2 = self._g2.nodes[j]

                    no_of_in_neighbors1 = self._g1.parent_count[n1]
                    no_of_in_neighbors2 = self._g2.parent_count[n2]
            
                    in_neighbor_sim = 0
                    if not max(no_of_in_neighbors1, no_of_in_neighbors2) == 0:
                        in_neighbors_matching_costs = [None] * no_of_in_neighbors1
                        for k in range(no_of_in_neighbors1):
                            in_neighbors_matching_costs[k] = [1] * no_of_in_neighbors2
                            for l in range(no_of_in_neighbors2):
                                #in_neighbors_matching_costs[k][l] = (1 - old_sim[n1.get_parent(k).index][n2.get_parent(l).index]) / self.__eps
                                in_neighbors_matching_costs[k][l] = ((
                                    1 - old_sim[self._g1.parent_node_idx(n1, k)][self._g2.parent_node_idx(n2, l)]
                                ) / self.__eps)
                                if self.__print_steps and i == 1 and j == 1:
                                    #print("in_neighbors_matching_costs[" + str(k) + "][" + str(l) + "] = (1 - old_sim[" + str(n1.get_parent(k).index) + "][" + str(n2.get_parent(l).index) + "]) / " + str(self.__eps))
                                    print("in_neighbors_matching_costs[" + str(k) + "][" + str(l) + "] = (1 - old_sim[" + str(self._g1.parent_node_idx(n1, k)) + "][" + str(self._g2.parent_node_idx(n2, l)) + "]) / " + str(self.__eps))
                                    #print("in_neighbors_matching_costs[" + str(k) + "][" + str(l) + "] = (1 -" + str(old_sim[n1.get_parent(k).index][n2.get_parent(l).index]) + ") / " + str(self.__eps))
                                    print("in_neighbors_matching_costs[" + str(k) + "][" + str(l) + "] = (1 -" + str(old_sim[self._g1.parent_node_idx(n1, k)][self._g2.parent_node_idx(n2, l)]) + ") / " + str(self.__eps))


                        #print("#### " + str(no_of_in_neighbors1) + " " + str(no_of_in_neighbors2) + " " + str(max(no_of_in_neighbors1, no_of_in_neighbors2)))
                        #print in_neighbors_matching_costs
                
                        # Print out the matrix for the paper
                        if self.__print_steps:
                            if i == 1 and j == 1:
                                print('Cost matrix for in neighbors of entry (2, 2):')
                                for k in range(no_of_in_neighbors1):
                                    for l in range(no_of_in_neighbors2):
                                        sys.stdout.write(str(in_neighbors_matching_costs[k][l]) + ' ')
                                    print('')
                
                        if no_of_in_neighbors1 > 0 and no_of_in_neighbors2 > 0:
                            munkres = Munkres()
                            indexes = munkres.compute(in_neighbors_matching_costs)
                            for row, column in indexes:
                                #value = old_sim[n1.get_parent(row).index][n2.get_parent(column).index]
                                value = old_sim[self._g1.parent_node_idx(n1, row)][self._g2.parent_node_idx(n2, column)]
                                if self.__print_steps and i == 1 and j == 1:
                                    #print(str(row) + ' ' + str(column) + ' ' + str(old_sim[n1.get_parent(row).index][n2.get_parent(column).index]))
                                    print('Match ' + str(row) + ' with ' + str(column) + '. Value: ' + str(old_sim[self._g1.parent_node_idx(n1, row)][self._g2.parent_node_idx(n2, column)]))
                                in_neighbor_sim += value
                            if self.__print_steps and i == 1 and j == 1:
                                print("s_in total: " + str(in_neighbor_sim))
                
                        in_neighbor_sim /= float(max(no_of_in_neighbors1, no_of_in_neighbors2))
                        if self.__print_steps and i == 1 and j == 1:
                            print("s_in / " + str(float(max(no_of_in_neighbors1, no_of_in_neighbors2))) + " = " + str(in_neighbor_sim))
                    else:
                        in_neighbor_sim = 1
                        if self.__print_steps and i == 1 and j == 1:
                            print("in_neighbor_sim = 1")
            
                    no_of_out_neighbors1 = self._g1.child_count[n1]
                    no_of_out_neighbors2 = self._g2.child_count[n2]
            
                    out_neighbor_sim = 0
                    if not max(no_of_out_neighbors1, no_of_out_neighbors2) == 0:
                        out_neighbors_matching_costs = [None] * no_of_out_neighbors1
                        for k in range(no_of_out_neighbors1):
                            out_neighbors_matching_costs[k] = [1] * no_of_out_neighbors2
                            for l in range(no_of_out_neighbors2):
                                #out_neighbors_matching_costs[k][l] = (1 - old_sim[n1.get_child(k).index][n2.get_child(l).index]) / self.__eps
                                out_neighbors_matching_costs[k][l] = (1 - old_sim[self._g1.child_node_idx(n1, k)][self._g2.child_node_idx(n2, l)]) / self.__eps

                        # Print out the matrix for the paper
                        if self.__print_steps and i == 1 and j == 1:
                            print('Cost matrix for out neighbors of entry (2, 2):')
                            for k in range(no_of_out_neighbors1):
                                for l in range(no_of_out_neighbors2):
                                    sys.stdout.write(str(out_neighbors_matching_costs[k][l]) + ' ')
                                print('')
                        
                        if self.__print_steps and i == 1 and j == 1:
                            print('Matching of out neighbors of entry (1, 1):')
                        if no_of_out_neighbors1 > 0 and no_of_out_neighbors2 > 0:
                            munkres = Munkres()
                            indexes = munkres.compute(out_neighbors_matching_costs)
                            for row, column in indexes:
                                #value = old_sim[n1.get_child(row).index][n2.get_child(column).index]
                                value = old_sim[self._g1.child_node_idx(n1, row)][self._g2.child_node_idx(n2, column)]
                                if self.__print_steps and i == 1 and j == 1:
                                    #print('Match ' + str(row) + ' with ' + str(column) + '. Value: ' + str(old_sim[n1.get_child(row).index][n2.get_child(column).index]))
                                    print('Match ' + str(row) + ' with ' + str(column) + '. Value: ' + str(old_sim[self._g1.child_node_idx(n1, row)][self._g2.child_node_idx(n2, column)]))
                                out_neighbor_sim += value
                            if self.__print_steps and i == 1 and j == 1:
                                print("s_out total: " + str(out_neighbor_sim))
                
                        out_neighbor_sim /= float(max(no_of_out_neighbors1, no_of_out_neighbors2))
                        if self.__print_steps and i == 1 and j == 1:
                            print("s_out / " + str(float(max(no_of_out_neighbors1, no_of_out_neighbors2))) + " = " + str(out_neighbor_sim))
                    else:
                        out_neighbor_sim = 1
            
                    #sim[i][j] = sqrt(n1.contrastByInstr(n2) * float(in_neighbor_sim + out_neighbor_sim) / 2)
                    sim[i][j] = float(in_neighbor_sim + out_neighbor_sim) / 2
            
                    if sim[i][j] - old_sim[i][j] >= self.__eps:
                        precise = False
    
            if precise:
                finished = True
    
            if self.__print_steps:
                print('Updated cost matrix:')
                for i in range(n):
                    for j in range(m):
                        sys.stdout.write(str(round(sim[i][j], 2)) + '\t')
                    print('')

        costs = [None] * n
        for i in range(n):
            costs[i] = [0] * m
            for j in range(m):
                costs[i][j] = float(1 - sim[i][j]) / self.__eps

        munkres = Munkres()
        indexes = munkres.compute(costs)

        matches = 0
        sim_score = 0
        for row, column in indexes:
            matches += 1
            sim_score += sim[row][column]

        if self._normalize:
            sim_score = float(sim_score) / matches

        return float(math.ceil(sim_score))


def hu_cfged(g1: nx.DiGraph, g2: nx.DiGraph, print_steps=False):
    ged = CFGSimNM(g1, g2, print_steps=print_steps)
    return ged.sim()



from collections import defaultdict
from typing import Optional

from ailment import BinaryOp, UnaryOp, Assignment
from ailment.block_walker import AILBlockWalkerBase
from ailment.block import Block
from ailment.statement import Call, ConditionalJump, Statement, Store, Return
from ailment.expression import Const, Load, ITE, Convert

from cfgutils.angr_utils.prettyify_ail import string_at_addr
from cfgutils.data.generic_statement import GenericStatement


class AILBlockConverter(AILBlockWalkerBase):
    def __init__(self, project=None, project_cfg=None):
        self.stmt_by_idx = defaultdict(list)
        self._project = project
        self._project_cfg = project_cfg
        super().__init__()

    @property
    def statements(self):
        ordered = []
        for idx, stmts in self.stmt_by_idx.items():
            ordered.append(reversed(stmts))

        return ordered

    def _fix_string_and_imms(self, const):
        if not isinstance(const, Const):
            return str(const)

        str_val = None
        if self._project is not None and self._project_cfg is not None:
            str_val = string_at_addr(self._project_cfg, const.value, self._project, max_size=200)

        return str_val if str_val is not None else const.value

    def _fix_list(self, _list):
        if not _list:
            return []

        out = []
        for _l in _list:
            out.append(self._fix_string_and_imms(_l))

        return out

    #
    # Statements
    #

    def _handle_Call(self, stmt_idx: int, stmt: "Call", block: Optional["Block"]):
        self.stmt_by_idx[stmt_idx].append(
            GenericStatement(stmt.tags['ins_addr'], "call", self._fix_list(stmt.args))
        )
        return super()._handle_Call(stmt_idx, stmt, block)

    def _handle_Return(self, stmt_idx: int, stmt: Return, block: Optional[Block]):
        self.stmt_by_idx[stmt_idx].append(
            GenericStatement(stmt.tags['ins_addr'], "ret", self._fix_list(stmt.ret_exprs))
        )
        return super()._handle_Return(stmt_idx, stmt, block)

    def _handle_Store(self, stmt_idx: int, stmt: Store, block: Optional[Block]):
        self.stmt_by_idx[stmt_idx].append(
            GenericStatement(stmt.tags['ins_addr'], "store", [
                self._fix_string_and_imms(stmt.addr), self._fix_string_and_imms(stmt.data)
            ])
        )
        return super()._handle_Store(stmt_idx, stmt, block)

    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Optional[Block]):
        self.stmt_by_idx[stmt_idx].append(
            GenericStatement(stmt.tags['ins_addr'], "assign", [
                self._fix_string_and_imms(stmt.dst), self._fix_string_and_imms(stmt.src)
            ])
        )
        return super()._handle_Assignment(stmt_idx, stmt, block)

    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Optional[Block]):
        self.stmt_by_idx[stmt_idx].append(
            GenericStatement(stmt.tags['ins_addr'], "cond_jmp", [
                self._fix_string_and_imms(stmt.condition), self._fix_string_and_imms(stmt.true_target),
                self._fix_string_and_imms(stmt.false_target)
            ])
        )
        return super()._handle_ConditionalJump(stmt_idx, stmt, block)

    #
    # Expr
    #

    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement, block: Optional[Block]):
        self.stmt_by_idx[stmt_idx].append(
            GenericStatement(stmt.tags['ins_addr'], "load", [
                self._fix_string_and_imms(expr.addr), self._fix_string_and_imms(expr.size)
            ])
        )
        return super()._handle_Load(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_ITE(self, expr_idx: int, expr: ITE, stmt_idx: int, stmt: Statement, block: Optional[Block]):
        self.stmt_by_idx[stmt_idx].append(
            GenericStatement(stmt.tags['ins_addr'], "ITE", [
                self._fix_string_and_imms(expr.cond), self._fix_string_and_imms(expr.iftrue),
                self._fix_string_and_imms(expr.iffalse)
            ])
        )
        return super()._handle_ITE(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
        self.stmt_by_idx[stmt_idx].append(
            GenericStatement(stmt.tags['ins_addr'], expr.op, self._fix_list(expr.operands))
        )
        return super()._handle_BinaryOp(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_UnaryOp(self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
        self.stmt_by_idx[stmt_idx].append(
            GenericStatement(stmt.tags['ins_addr'], expr.op, self._fix_string_and_imms(expr.operand))
        )
        return super()._handle_UnaryOp(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_CallExpr(self, expr_idx: int, expr: "Call", stmt_idx: int, stmt, block: Optional["Block"]):
        self.stmt_by_idx[stmt_idx].append(
            GenericStatement(stmt.tags['ins_addr'], "call", self._fix_list(expr.args))
        )
        return super()._handle_CallExpr(expr_idx, expr, stmt_idx, stmt, block)




import hashlib
from collections import defaultdict
from typing import Dict, Optional, List, Tuple

from .flowgraph import FlowGraph
from . import rotl64, mask64bit, seed0_, seed1_, seed2_, k0, k1, k2, mask32bit


class FunctionSimHasher:
    TYP_GRAPH = "graph"
    TYPE_MNEM = "mnemonic"
    TYPE_IMM = "immediate"

    def __init__(self, w_graph=1.0, w_mnem=0.05, w_imm=4.0):
        # weights
        self.w_graph = w_graph
        self.w_mnem = w_mnem
        self.w_imm = w_imm

    def CalculateFunctionSimHash(self, graph, bit_size=128):
        if bit_size % 64 != 0:
            raise ValueError("Requested hash bit size must be a multiple of 64!")

        full_flow_graph = FlowGraph(graph)
        feature_cardnalities = defaultdict(int)
        final_floats = [0.0]*bit_size

        # graphlets (sub-graphs)
        for node, dist in full_flow_graph.nodes_and_distance:
            graphlet = full_flow_graph.GetSubgraph(graph, node, dist)
            if graphlet is None:
                continue

            _id = self._hash_graph(graphlet, node, hash_index=0, counter=0)
            card = feature_cardnalities[_id]
            feature_cardnalities[_id] += 1
            #feat_card_id = self._hash_graph(graphlet, node, card, 0)
            # XXX: since we dont support the getWeight() function, its just the weight of that feature
            weight = self.w_graph

            # start ProcessSubgraph here
            # calculate nbithash
            _hashes = [
                self._hash_graph(graphlet, node, hash_index=card, counter=(cntr + 1) * 64)
                for cntr in range(bit_size // 64)
            ]
            # skip adding to feature hashes
            self.AddWeightsInHashToOutput(final_floats, bit_size, weight, _hashes)

        # mnemonics (instruction operations)
        for mnem_tup in full_flow_graph.mnemonic_ngrams:
            _id = self._hash_mnemonic_tuple(mnem_tup, hash_index=0)
            card = feature_cardnalities[_id]
            feature_cardnalities[_id] += 1
            weight = self.w_mnem
            _hashes = [
                # TODO: might want to verify this is correct for hash_index
                self._hash_mnemonic_tuple(mnem_tup, hash_index=card + (((cntr + 1) * 64) + 1))
                for cntr in range(bit_size // 64)
            ]
            self.AddWeightsInHashToOutput(final_floats, bit_size, weight, _hashes)

        # immediates
        for imm in full_flow_graph.immediates:
            _id = self._hash_immediate(imm, hash_index=0, counter=0)
            card = feature_cardnalities[_id]
            feature_cardnalities[_id] += 1
            weight = self.w_imm
            _hashes = [
                self._hash_immediate(imm, hash_index=card, counter=(cntr + 1) * 64)
                for cntr in range(bit_size // 64)
            ]
            self.AddWeightsInHashToOutput(final_floats, bit_size, weight, _hashes)

        vals = self.FloatsToBits(final_floats)
        return vals[:-1]

    def AddWeightsInHashToOutput(self, final_floats, bit_size, weight, hashes):
        """
        Updates final_floats in place
        """
        for bit_n in range(bit_size):
            if self.GetNthBit(hashes, bit_n):
                final_floats[bit_n] += weight
            else:
                final_floats[bit_n] -= weight

    def _stable_hash(self, string: str) -> int:
        b_str = string.encode()
        return int(hashlib.md5(b_str).hexdigest(), 16)

    def _hash_mnemonic_tuple(self, mnem_tuple: Tuple[str, str, str], hash_index=0):
        m0, m1, m2 = mnem_tuple
        value1 = (
            self.SeedXForHashY(0, hash_index) ^ self.SeedXForHashY(1, hash_index) ^
            self.SeedXForHashY(2, hash_index)
        ) & mask64bit
        value1 *= self._stable_hash(m0) & mask64bit
        value1 = rotl64(value1, 7)
        value1 *= self._stable_hash(m1) & mask64bit
        value1 = rotl64(value1, 7)
        value1 *= self._stable_hash(m2) & mask64bit
        value1 = rotl64(value1, 7)
        value1 *= (k2 * (hash_index + 1)) & mask64bit
        return value1 & mask64bit

    def _hash_immediate(self, immediate, hash_index=0, counter=0):
        value1 = (
            self.SeedXForHashY(0, hash_index) & mask64bit +
            (counter * k0) & mask64bit +
            (counter * k1) & mask64bit +
            (counter * k2) & mask64bit
        ) & mask64bit
        value1 = rotl64(value1, 7)
        value1 *= (immediate ^ self.SeedXForHashY(0, hash_index)) & mask64bit
        value1 &= mask64bit
        value1 = rotl64(value1, 7)
        value1 *= (immediate ^ self.SeedXForHashY(1, hash_index)) & mask64bit
        value1 &= mask64bit
        value1 = rotl64(value1, 7)
        value1 *= (immediate ^ self.SeedXForHashY(2, hash_index)) & mask64bit
        value1 &= mask64bit
        value1 = rotl64(value1, 7)
        value1 *= ((k2 ^ immediate) * (hash_index + 1)) & mask64bit
        value1 &= mask64bit
        return value1

    def _hash_graph(self, flow_graph: FlowGraph, start_node, hash_index=0, counter=0):
        return flow_graph.CalculateHash(
            start_node=start_node,
            k0=self.SeedXForHashY(0, hash_index) * (counter + 1),
            k1=self.SeedXForHashY(1, hash_index) * (counter + 1),
            k2=self.SeedXForHashY(2, hash_index) * (counter + 1)
        )

    #
    # Bit Utils
    #

    @staticmethod
    def FloatsToBits(floats: List[float]):
        outputs = [0] * ((len(floats) // 64) + 1)
        for index, _float in enumerate(floats):
            vector_index = index // 64
            bit_index = index % 64
            if _float >= 0:
                outputs[vector_index] |= (1 << bit_index)

        return outputs

    @staticmethod
    def GetNthBit(nbit_hash: List[int], bitindex):
        index = bitindex // 64
        value = nbit_hash[index]
        sub_word_index = bitindex % 64
        return (value >> sub_word_index) & 1

    @staticmethod
    def SeedXForHashY(seed_index, hash_index):
        if seed_index == 0:
            return rotl64(seed0_, hash_index % 7) * (hash_index + 1)
        elif seed_index == 1:
            return rotl64(seed1_, hash_index % 11) * (hash_index + 1)
        elif seed_index == 2:
            return rotl64(seed2_, hash_index % 13) * (hash_index + 1)
        else:
            raise ValueError("Invalid seed index")

    @staticmethod
    def hash_distance(h1: List[int], h2: List[int]):
        total_dist = 0
        for _h1, _h2 in zip(h1, h2):
            total_dist += _h1 ^ _h2

        return total_dist


# this entire files structure should only be imported if you have angr installed in the same
# environment as this package.
#
from .ail_graph_conv import binary_to_ail_cfgs, binary_to_generic_cfgs, ail_pickle_to_cfg, ail_cfg_to_generic



from typing import List, Optional


class GenericBlock:
    def __init__(
        self, addr: int, statements: List = None, idx: Optional[int] = None, is_entrypoint: bool = False,
        is_exitpoint: bool = False, is_merged_node=False
    ):
        self.addr = addr
        self.statements = statements or []
        self.idx = idx
        self.is_entrypoint = is_entrypoint
        self.is_exitpoint = is_exitpoint
        self.is_merged_node = is_merged_node

        self._idx_str = "" if self.idx is None else f".{self.idx}"

    def __eq__(self, other):
        return type(other) is self.__class__ and self.addr == other.addr and self.statements == other.statements

    def __hash__(self):
        return hash(f"{self.addr}{self.idx}{[d for d in self.statements]}")

    def __repr__(self):
        type_str = " (exit)" if self.is_exitpoint else " (entry)" if self.is_entrypoint else ""
        return f"<Block: {hex(self.addr)}{self._idx_str}{type_str}>"

    def __str__(self):
        type_str = " (exit)" if self.is_exitpoint else " (entry)" if self.is_entrypoint else ""
        output = f"{hex(self.addr)}{self._idx_str}{type_str}:\n"
        for stmt in self.statements:
            str_stmt = str(stmt)
            if len(str_stmt) > 80:
                str_stmt = str_stmt[:80] + "..."
            output += f"{str_stmt}\n"
        return output

    def copy(self):
        return self.__class__(
            self.addr,
            statements=self.statements.copy() if self.statements is not None else None,
            idx=self.idx,
        )

    def contains_addr(self, addr):
        if self.addr == addr:
            return True

    @classmethod
    def merge_blocks(cls, block1: "GenericBlock", block2: "GenericBlock"):
        new_node = block1.copy()
        new_node.statements += block2.statements
        new_node.is_entrypoint |= block2.is_entrypoint
        new_node.is_exitpoint |= block2.is_exitpoint
        new_node.is_merged_node = True
        return new_node

    @classmethod
    def merge_many_blocks(cls, start_addr, nodes: List["GenericBlock"]):
        new_node = nodes[0].copy()
        new_node.addr = start_addr
        for node in nodes[1:]:
            new_node = cls.merge_blocks(new_node, node)

        return new_node



import networkx as nx

from .data.generic_block import GenericBlock


def merge_graph_nodes(graph: nx.DiGraph, node_a: GenericBlock, node_b: GenericBlock):
    in_edges = list(graph.in_edges(node_a, data=True))
    out_edges = list(graph.out_edges(node_b, data=True))
    new_node = node_a.merge_blocks(node_a, node_b)

    graph.remove_node(node_a)
    graph.remove_node(node_b)

    if new_node is not None:
        graph.add_node(new_node, node=new_node)

        for src, _, data in in_edges:
            if src is node_b:
                src = new_node
            graph.add_edge(src, new_node, src=src, dst=new_node)

        for _, dst, data in out_edges:
            if dst is node_a:
                dst = new_node
            graph.add_edge(new_node, dst, src=new_node, dst=dst)

    return new_node


def to_supergraph(graph: nx.DiGraph):
    new_graph = nx.DiGraph(graph)
    while True:
        for src, dst in new_graph.edges():
            if len(list(new_graph.successors(src))) == 1 and len(list(new_graph.predecessors(dst))) == 1:
                if src is not dst:
                    merge_graph_nodes(new_graph, src, dst)
                    break
        else:
            break

    return new_graph



import itertools
from collections import defaultdict
from typing import List, Dict, Tuple, Union

import networkx as nx

from ..sorting import quasi_topological_sort_nodes


class BlockMatcherBase:
    """
    This class takes two graphs and returns a dictionary of likely matches between the nodes in the two graphs.
    The key is nodes in g1 and the value is the corresponding node in g2. This class is expected to be used
    with two graphs you _know_ are implementing the same logic. This can happen in instances of a function being
    compiled with different optimization levels, or with different compilers.

    The core score of each block is based on the paper: "discovRE: Efficient Cross-Architecture Identification of Bugs
    in Binary Code" by Eschweiler et al. in NDSS 2016. After collecting these initial scores, we will do more analysis
    to get more matches.
    """
    W_NO_ARITHMETIC_INS = 56.658
    W_NO_CALLS = 87.423
    W_NO_INS = 40.423
    W_NO_LOGIC_INS = 76.694
    W_NO_BRANCH_INS = 6.841
    W_STR_CONST = 11.998
    W_NUM_CONST = 15.382

    # guessed weights
    W_TOPO_DIST = 15.0  # as important a number matching across blocks

    TYP_NO_ARITHMETIC_INS = "arith_ins"
    TYP_NO_CALLS = "calls"
    TYP_NO_INS = "ins"
    TYP_NO_LOGIC_INS = "logic_ins"
    TYP_NO_BRANCH_INS = "branch_ins"
    TYP_STR_CONST = "str_consts"
    TYP_NUM_CONST = "num_consts"
    TYP_TOPO_DIST = "topo_dist"

    def __init__(
        self,
        g1: nx.DiGraph,
        g2: nx.DiGraph,
        use_new_weights=False,
        use_inv_map=False,
        assume_rooted=True,
        graph_match_pass=False,
        match_confidence=0.0,
        root_dist_tie_breaker=True,
        mapping=None,
        out_edges_must_match=True,
    ):
        self._g1 = g1
        self._g2 = g2
        self._use_new_weights = use_new_weights
        self._use_inv_map = use_inv_map
        self._graph_match_pass = graph_match_pass
        self._assume_rooted = assume_rooted
        self._root_dist_tie_breaker = root_dist_tie_breaker
        self._out_edges_must_match = out_edges_must_match
        self.match_confidence = match_confidence

        self._g1_nodes = list(g1.nodes)
        self._g2_nodes = list(g2.nodes)
        self._topo_dist = {
            g1: self._generate_topological_dist_map(g1),
            g2: self._generate_topological_dist_map(g2)
        }

        self._g1_root, self._g2_root = None, None
        if self._assume_rooted:
            self._g1_root = next(n for n in self._g1_nodes if self._g1.in_degree(n) == 0)
            self._g2_root = next(n for n in self._g2_nodes if self._g2.in_degree(n) == 0)

        self._root_dist = {}
        if self._root_dist_tie_breaker and self._assume_rooted:
            self._root_dist = {
                g1: nx.single_source_shortest_path_length(g1, self._g1_root),
                g2: nx.single_source_shortest_path_length(g2, self._g2_root),
            }

        self._used_b2 = set()
        self.mapping = mapping or {}

    def neighbour_match_nodes(self, g1: nx.DiGraph, g2: nx.DiGraph):
        changes = True
        max_attempts = 1000
        attempts = 0
        while changes and attempts < max_attempts:
            changes = False
            attempts += 1
            chosen_b2s = set(self.mapping.values())

            for b1, b2 in list(self.mapping.items()):
                uniq_b1_preds = list(pred for pred in g1.predecessors(b1) if pred not in self.mapping)
                uniq_b2_preds = list(pred for pred in g2.predecessors(b2) if pred not in chosen_b2s)

                # Case1: a matched set both have one parent not matched yet
                if len(uniq_b1_preds) == 1 and len(uniq_b2_preds) == 1:
                    self.mapping[uniq_b1_preds[0]] = uniq_b2_preds[0]
                    changes = True
                    break

                uniq_b1_succs = list(succ for succ in g1.successors(b1) if succ not in self.mapping)
                uniq_b2_succs = list(succ for succ in g2.successors(b2) if succ not in chosen_b2s)
                # Case2: a matched set both have one child
                if len(uniq_b1_succs) == 1 and len(uniq_b2_succs) == 1:
                    self.mapping[uniq_b1_succs[0]] = uniq_b2_succs[0]
                    changes = True
                    break

        return attempts > 1

    def analyze(self):
        # mark all nodes provided already as matches
        self._used_b2 = set()
        for _, g2_node in self.mapping.items():
            self._used_b2.add(g2_node)

        b1_b2_map, best_b2_scores = self.generate_similarities(get_best_inv_map=True)
        if self._assume_rooted:
            self.mapping[self._g1_root] = self._g2_root
            self._used_b2.add(self._g2_root)

            g1_exits = [n for n in self._g1_nodes if self._g1.out_degree(n) == 0]
            g2_exits = [n for n in self._g2_nodes if self._g2.out_degree(n) == 0]
            if len(g1_exits) == len(g2_exits) == 1:
                self.mapping[g1_exits[0]] = g2_exits[0]
                self._used_b2.add(g2_exits[0])

        if self._use_inv_map:
            # first, map every b2 to the b1 that matches it best
            for b2, b1 in best_b2_scores.items():
                if (b2 in self._used_b2) or (b1 in self.mapping):
                    continue

                self.mapping[b1] = b2
                self._used_b2.add(b2)

        # then, map the rest of the b1s to the b2s that match them best
        for b1, b2_scores in b1_b2_map.items():
            if b1 in self.mapping:
                continue

            best_matches = sorted(b2_scores.items(), key=lambda x: x[1], reverse=True)
            best_score = best_matches[0][1]
            top_score_matches = [b2 for b2, score in best_matches if score == best_score]
            # eliminate missing edge matches
            if self._out_edges_must_match:
                top_score_matches = [
                    b2 for b2 in top_score_matches if set(self._g1.successors(b1)) == set(self._g2.successors(b2))
                ]

            # if tie-break enable, attempt to resolve the top-level ties (anything after is ok...)
            if len(top_score_matches) > 1 and self._root_dist_tie_breaker:
                b1_dist = self._root_dist[self._g1][b1]
                _best_first_matches = sorted(
                    best_matches[:len(top_score_matches)], key=lambda x: abs(b1_dist - self._root_dist[self._g2][x[0]])
                )
                best_matches = _best_first_matches + best_matches[len(top_score_matches):]

            for blk_match, score in best_matches:
                # scores are ordered, so if the top score is below the confidence, we can break
                if score < self.match_confidence:
                    break

                if blk_match not in self._used_b2:
                    self.mapping[b1] = blk_match
                    self._used_b2.add(blk_match)
                    break
            else:
                # if no match was found at all, we must be out of b2s
                break

        if self._graph_match_pass:
            self.neighbour_match_nodes(self._g1, self._g2)

    def generate_similarities(self, get_best_inv_map=False) -> \
            Union[Tuple[Dict[object, Dict[object, float]], Dict[object, float]], Dict[object, Dict[object, float]]]:
        b1_to_b2_scores = defaultdict(dict)
        best_b2_scores = {}
        for b1, b2 in itertools.product(self._g1_nodes, self._g2_nodes):
            b1_to_b2 = self.block_similarity(b1, b2, self._g1, self._g2)
            b1_to_b2_scores[b1][b2] = b1_to_b2
            best_b2_score, _ = best_b2_scores.get(b2, (0, None))
            if b1_to_b2 > best_b2_score:
                best_b2_scores[b2] = (b1_to_b2, b1)

        best_b2_scores = {k: v[1] for k, v in best_b2_scores.items()}
        if get_best_inv_map:
            return b1_to_b2_scores, best_b2_scores
        else:
            return b1_to_b2_scores

    def block_similarity(self, b1, b2, g1, g2) -> float:
        feat_func_map = {
            self.get_number_of_arithmetic_ins: self.W_NO_ARITHMETIC_INS,
            self.get_number_of_calls: self.W_NO_CALLS,
            self.get_number_of_ins: self.W_NO_INS,
            self.get_number_of_logic_ins: self.W_NO_LOGIC_INS,
            self.get_number_of_branch_ins: self.W_NO_BRANCH_INS,
        }
        listed_feat_func_map = {
            self.get_str_consts: self.W_STR_CONST,
            self.get_num_consts: self.W_NUM_CONST
        }
        # use weights that are not from the paper, but might be useful
        if self._use_new_weights:
            feat_func_map[self.get_topological_dist] = self.W_TOPO_DIST

        numerator = 0
        denominator = 0
        # features that are a single value
        for feat_func, weight in feat_func_map.items():
            f1, f2 = feat_func(b1, g1), feat_func(b2, g2)
            numerator += weight * abs(f1 - f2)
            denominator += weight * max(f1, f2)

        # features that are a list
        for feat_func, weight in listed_feat_func_map.items():
            f1, f2 = feat_func(b1, g1), feat_func(b2, g2)
            # jaccard similarity
            numerator += (len(set.union(set(f1), set(f2))) - len(set.intersection(set(f1), set(f2)))) * weight
            denominator += len(set.union(set(f1), set(f2))) * weight

        # denominator == 0 when two empty blocks compared, they are equal
        dissimilarity = (numerator / denominator) if denominator else 0
        return 1 - dissimilarity

    @staticmethod
    def _generate_topological_dist_map(graph: nx.DiGraph) -> Dict[object, int]:
        return {
            node: dist for dist, node in enumerate(quasi_topological_sort_nodes(graph))
        }

    def get_topological_dist(self, block, graph: nx.DiGraph):
        return self._topo_dist[graph][block]

    #
    # Must be implemented by the user
    #

    def get_number_of_arithmetic_ins(self, block, graph) -> int:
        raise NotImplementedError()

    def get_number_of_calls(self, block, graph) -> int:
        raise NotImplementedError()

    def get_number_of_ins(self, block, graph) -> int:
        raise NotImplementedError()

    def get_number_of_logic_ins(self, block, graph) -> int:
        raise NotImplementedError()

    def get_number_of_branch_ins(self, block, graph) -> int:
        raise NotImplementedError()

    def get_str_consts(self, block, graph) -> List[str]:
        raise NotImplementedError()

    def get_num_consts(self, block, graph) -> List[int]:
        raise NotImplementedError()



import sys
import unittest

from cfgutils.regions.region_identifier import RegionIdentifier
from cfgutils.data import numbered_edges_to_block_graph


class TestRegionIdentification(unittest.TestCase):
    def test_region_identification(self):
        #
        # Graph:
        #         1
        #        / \
        #       2   3
        #       \   /
        #         4
        #        / \
        #       5   6
        #        \ /
        #         7
        #
        # Regions:
        # [1,2,3]
        # [4,5,6]
        # [1,7]
        # None blocks added to start and end to make indexing easier
        graph = numbered_edges_to_block_graph(
            [(1, 2), (1, 3), (2, 4), (3, 4), (4, 5), (4, 6), (5, 7), (6, 7)]
        )
        ri = RegionIdentifier(graph)
        top_region = ri.region

        # the node to start all the regions should be 1
        assert top_region.head.head.head.addr == 1

        # check known regions
        region_blk_sets = [set(blks) for blks in ri.regions_by_block_addrs]
        assert {1, 2, 3} in region_blk_sets
        assert {4, 5, 6} in region_blk_sets
        assert {1, 7} in region_blk_sets


if __name__ == "__main__":
    unittest.main(argv=sys.argv)


import sys
import unittest

from cfgutils.matrix.munkres import Munkres, print_matrix


class TestMatrixMath(unittest.TestCase):
    def test_munkres(self):
        matrices = [
            # Square
            ([[400, 150, 400],
              [400, 450, 600],
              [300, 225, 300]],
             850  # expected cost
             ),

            # Rectangular variant
            ([[400, 150, 400, 1],
              [400, 450, 600, 2],
              [300, 225, 300, 3]],
             452  # expected cost
             ),

            # Square
            ([[10, 10, 8],
              [9, 8, 1],
              [9, 7, 4]],
             18
             ),

            # Rectangular variant
            ([[10, 10, 8, 11],
              [9, 8, 1, 1],
              [9, 7, 4, 10]],
             15
             ),
        ]

        m = Munkres()
        for cost_matrix, expected_total in matrices:
            print_matrix(cost_matrix, msg='cost matrix')
            indexes = m.compute(cost_matrix)
            total_cost = 0
            for r, c in indexes:
                x = cost_matrix[r][c]
                total_cost += x
                print('(%d, %d) -> %d' % (r, c, x))
            print('lowest cost=%d' % total_cost)
            assert expected_total == total_cost


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



# Based on the default GED algorithm found in Networkx which is based on the following paper:
# "An Exact Graph Edit Distance Algorithm for Solving Pattern Recognition Problems" by Abu-Aisheh et al.
#

import logging
import os

import networkx as nx

from cfgutils.os_utils import timeout
from cfgutils.similarity import fast_is_isomorphic
from cfgutils.similarity.ged import INVALID_CHOICE_PENALTY, collect_graph_roots

_l = logging.getLogger(__name__)
MAX_NODES_FOR_EXACT_GED = 12
MIN_EXACT_TIME = 4
MIN_UPPERBOUND_TIME = 3

_DEBUG = bool(os.getenv("DEBUG", False)) or False
if _DEBUG:
    _l.setLevel(logging.DEBUG)

#
# Edit distance
#

def ged_max(g1, g2):
    return len(g1.nodes) + len(g1.edges) + len(g2.nodes) + len(g2.edges)


def ged_exact(g1, g2, with_timeout=10, check_max=False):
    """
    Computes the exact Graph Edit Distance for two graphs. On the event of a timeout,
    a score of None is returned.
    """
    if fast_is_isomorphic(g1, g2):
        return 0

    if check_max and (len(g1.nodes) > MAX_NODES_FOR_EXACT_GED or len(g2.nodes) > MAX_NODES_FOR_EXACT_GED):
        return None

    return graph_edit_distance_core_analysis(g1, g2, with_timeout=with_timeout, exact_score=True)


def ged_upperbound(g1, g2, with_timeout=5):
    """
    Does a single iterations of the GED algorithm and returns the upperbound.
    Note: this is not the max possible score.
    """
    return graph_edit_distance_core_analysis(g1, g2, upperbound_approx=True, with_timeout=with_timeout)


def ged_explained(g1, g2, print_explanation=True, only_addrs=True):
    """

    Possible operations:
    ins(node)
    del(node)
    ins((node1, node2))
    del((node1, node2))
    """
    paths, cost = nx.optimal_edit_paths(g1, g2, node_subst_cost=_cfg_node_sub_cost, edge_subst_cost=_cfg_edge_sub_cost)
    chosen_path = paths[0]
    vertex_edits = chosen_path[0]
    edge_edits = chosen_path[1]
    human_v_edits = []
    human_e_edits = []
    human_v_swaps = []

    for v_edit in vertex_edits:
        v1, v2 = v_edit
        if v1 is None and v2 is not None:
            human_v_edits.append(f"ins({v2.__repr__() if not only_addrs else v2.addr})")
        elif v1 is not None and v2 is None:
            human_v_edits.append(f"del({v1.__repr__() if not only_addrs else v1.addr})")
        elif v2 is not None and v1 is not None:
            human_v_swaps.append(f"{v1.__repr__()} = {v2.__repr__()}")

    for e_edit in edge_edits:
        e1, e2 = e_edit
        if e1 is None and e2 is not None:
            e_str = f"ins({e2})" if not only_addrs else f"ins({e2[0].addr}, {e2[1].addr})"
            human_e_edits.append(e_str)
        elif e1 is not None and e2 is None:
            e_str = f"del({e1})" if not only_addrs else f"del({e1[0].addr}, {e1[1].addr})"
            human_e_edits.append(e_str)
        else:
            pass

    # sort explanations by del always being first
    human_v_edits = list(sorted(human_v_edits, key=lambda x: x.startswith("del"), reverse=True))
    human_e_edits = list(sorted(human_e_edits, key=lambda x: x.startswith("del"), reverse=True))

    if print_explanation:
        if cost:
            print("\n=====================================================================")
            print(f"There are {len(paths)} possible edit strategies to turn g1 into g2")
            print(f"in {cost} steps. Showing the first one:")
            print("=====================================================================")

            if human_v_swaps:
                print("Vertex Swaps (free): ")
                for v_swap in human_v_swaps:
                    print(f"    {v_swap}")
            print("Vertex Edits:")
            for v_edit in human_v_edits:
                print(f"    {v_edit}")
            print("Edge Edits:")
            for e_edit in human_e_edits:
                print(f"    {e_edit}")
            print("=====================================================================")
        else:
            print("No cost to explain. All edits were swaps.")

    return human_v_edits + human_e_edits, cost

#
# Updates to the original GED algorithm
#


def _cfg_node_sub_cost(*args, penalize_root_exit_edits=True):
    """
    Makes it illegal to delete function start nodes or end nodes
    """
    node_attrs = args[:2]
    n1, n2 = node_attrs[0].get('node', None), node_attrs[1].get('node', None)
    # 0 is the normal cost of a substitution that is valid, so we only update if we tried an illegal
    # subst on starts or exits
    cost = 0
    if n1 is None or n2 is None:
        print("HIT A CASE OF SOMEONE BEING NONE!")

    if penalize_root_exit_edits and (n1 and n2):
        if n1.is_entrypoint and not n2.is_entrypoint:
            cost = INVALID_CHOICE_PENALTY
        elif n1.is_exitpoint and not n2.is_exitpoint:
            cost = INVALID_CHOICE_PENALTY
        elif n2.is_entrypoint and not n1.is_entrypoint:
            cost = INVALID_CHOICE_PENALTY
        elif n2.is_exitpoint and not n1.is_exitpoint:
            cost = INVALID_CHOICE_PENALTY

    return cost


def _cfg_edge_sub_cost(*args, penalize_root_exit_edits=True):
    edge_attrs = args[:2]
    s1, d1 = edge_attrs[0].get('src', None), edge_attrs[0].get('dst', None)
    s2, d2 = edge_attrs[1].get('src', None), edge_attrs[1].get('dst', None)
    cost = 0
    if penalize_root_exit_edits and s1 and d1 and s2 and d2:
        node_pairs = {
            s1: s2,
            s2: s1,
            d1: d2,
            d2: d1,
        }

        for n1, n2 in node_pairs.items():
            if n1.is_entrypoint and not n2.is_entrypoint:
                cost = INVALID_CHOICE_PENALTY
            elif n1.is_exitpoint and not n2.is_exitpoint:
                cost = INVALID_CHOICE_PENALTY
            elif n2.is_entrypoint and not n1.is_entrypoint:
                cost = INVALID_CHOICE_PENALTY
            elif n2.is_exitpoint and not n1.is_exitpoint:
                cost = INVALID_CHOICE_PENALTY

            # early exit
            if cost != 0:
                break

    return cost


def graph_edit_distance_core_analysis(
    g1: nx.DiGraph, g2: nx.DiGraph, is_cfg=True, upperbound_approx=False, exact_score=False, with_timeout=10,
    penalize_root_exit_edits=True, recover_on_invalid_edits=True
):
    if fast_is_isomorphic(g1, g2):
        return 0

    roots = collect_graph_roots(g1, g2) if is_cfg else None

    _node_sub_cost = None
    _edge_sub_cost = None
    if is_cfg and penalize_root_exit_edits:
        _node_sub_cost = _cfg_node_sub_cost
        _edge_sub_cost = _cfg_edge_sub_cost

    if exact_score or upperbound_approx:
        try:
            with timeout(seconds=with_timeout):
                if upperbound_approx:
                    dist = next(nx.optimize_graph_edit_distance(
                        g1, g2, node_subst_cost=_node_sub_cost, edge_subst_cost=_edge_sub_cost,
                    ))
                else:
                    dist = nx.graph_edit_distance(
                        g1, g2, roots=roots, node_subst_cost=_node_sub_cost, edge_subst_cost=_edge_sub_cost,
                    )
        except TimeoutError:
            dist = None
    else:
        if _DEBUG:
            edit_path, dist = ged_explained(
                g1, g2, print_explanation=True, only_addrs=False
            )
        else:
            dist = nx.graph_edit_distance(
                g1, g2, roots=roots, timeout=with_timeout, node_subst_cost=_node_sub_cost, edge_subst_cost=_edge_sub_cost,
            )

    # sometimes the score can be computed wrong, which we can fix with a recompute ONCE
    if dist is not None and dist >= INVALID_CHOICE_PENALTY and recover_on_invalid_edits:
        dist = graph_edit_distance_core_analysis(
            g1, g2, is_cfg=is_cfg, upperbound_approx=upperbound_approx, exact_score=exact_score,
            with_timeout=with_timeout, penalize_root_exit_edits=False, recover_on_invalid_edits=False
        )

    return dist



from typing import Union
from pathlib import Path

import networkx as nx
import graphviz


def save_cfg_as_png(cfg: nx.DiGraph, output_path: Union[Path, str]):
    output_path = Path(output_path)
    tmp_path = output_path.with_suffix(".dot")
    nx.drawing.nx_agraph.write_dot(cfg, str(tmp_path))
    dot_src = graphviz.Source(open(tmp_path).read(), format="png")
    dot_src.render(outfile=str(output_path.with_suffix(".png")))
    tmp_path.with_suffix(".gv").unlink()
    tmp_path.unlink()
    return output_path.with_suffix(".png")



from collections import defaultdict
from typing import List

import networkx as nx
from ailment import Block
from ailment.statement import Call

from cfgutils.sorting import quasi_topological_sort_nodes
from cfgutils.similarity.block_matcher_base import BlockMatcherBase
from .feat_extractor import AILBlockFeatureExtractor


class AILBlockMatcher(BlockMatcherBase):
    TYP_VAR_NAMES = "var_names"
    TYP_STACK_ADDRS = "stack_addrs"
    TYP_FUNC_NAMES = "func_names"

    def __init__(
        self,
        g1: nx.DiGraph,
        g2: nx.DiGraph,
        proj1=None,
        proj2=None,
        match_confidence=0.6,
        graph_match_pass=True,
        mapping=None,
        use_var_names=True,
        use_caller_names=True,
        caller_name_mapping=None,
        root_dist_tie_breaker=True,
        match_exact_calls=True,
        assume_rooted=True,
    ):
        super().__init__(
            g1,
            g2,
            match_confidence=match_confidence,
            graph_match_pass=graph_match_pass,
            mapping=mapping,
            root_dist_tie_breaker=root_dist_tie_breaker,
            assume_rooted=assume_rooted,
        )
        self._proj1 = proj1
        self._proj2 = proj2
        self._use_var_names = use_var_names
        self._use_caller_names = use_caller_names
        self._caller_name_mapping = caller_name_mapping or {}

        self._g1_cache = defaultdict(dict)
        self._g2_cache = defaultdict(dict)
        self.generate_feat_cache()

        if match_exact_calls:
            self.mapping = self.match_exact_calls()

        self.analyze()

    def match_exact_calls(self):
        """
        Finds all calls which are exactly the same with the same root distance.

        :return:
        """
        mapping = {}
        for b1 in self._g1.nodes:
            choices = []
            # search all cases of names and amount lining up
            for b2 in self._g2.nodes:
                if (
                    self.get_number_of_calls(b1, self._g1) == self.get_number_of_calls(b2, self._g2) != 0 and
                    # XXX: this could be bad, order should matter...
                    set(self._g1_cache[b1][self.TYP_FUNC_NAMES]) == set(self._g2_cache[b2][self.TYP_FUNC_NAMES])
                ):
                    choices.append(b2)

            # since there can be multiple blocks that meet the above criteria, if we have to choose,
            # then we must be sure its the same. To do that, we use their distance from the root.
            if len(choices) == 1:
                mapping[b1] = choices[0]
            else:
                # first attempt to eliminate all choices with different edge amounts
                same_edge_choices = [
                    choice for choice in choices
                    if len(list(self._g2.predecessors(choice))) == len(list(self._g1.predecessors(b1))) and
                    len(list(self._g2.successors(choice))) == len(list(self._g1.successors(b1)))
                ]
                if len(same_edge_choices) == 1:
                    mapping[b1] = same_edge_choices[0]
                # if we still have multiple choices, we must use root distance to know truth
                # go back to before and use root distance to choose
                else:
                    for choice in choices:
                        if self._root_dist[self._g1][b1] == self._root_dist[self._g2][choice]:
                            mapping[b1] = choice
                            break

        return mapping

    def generate_feat_cache(self):
        for g, proj, features in [(self._g1, self._proj1, self._g1_cache), (self._g2, self._proj2, self._g2_cache)]:
            proj_cfg = proj.kb.cfgs.get_most_accurate() if proj is not None else None
            for node in g.nodes:
                feat_extractor = AILBlockFeatureExtractor(proj, proj_cfg, call_name_fallback=self._caller_name_mapping)
                feat_extractor.walk(node)
                features[node][self.TYP_NO_ARITHMETIC_INS] = len(feat_extractor.arith_ins)
                features[node][self.TYP_NO_CALLS] = len(feat_extractor.calls)
                features[node][self.TYP_NO_INS] = len(feat_extractor.ins)
                features[node][self.TYP_NO_LOGIC_INS] = len(feat_extractor.logic_ins)
                features[node][self.TYP_NO_BRANCH_INS] = len(feat_extractor.branch_ins)
                features[node][self.TYP_STR_CONST] = feat_extractor.str_consts
                features[node][self.TYP_NUM_CONST] = feat_extractor.num_consts
                features[node][self.TYP_VAR_NAMES] = feat_extractor.var_names
                features[node][self.TYP_STACK_ADDRS] = feat_extractor.stack_addrs
                features[node][self.TYP_FUNC_NAMES] = feat_extractor.call_names

    def _get_correct_cache(self, graph):
        if graph == self._g1:
            return self._g1_cache
        elif graph == self._g2:
            return self._g2_cache
        else:
            raise ValueError("Graph not found")

    def get_number_of_arithmetic_ins(self, block, graph) -> int:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NO_ARITHMETIC_INS]

    def get_number_of_calls(self, block, graph) -> int:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NO_CALLS]

    def get_number_of_ins(self, block, graph) -> int:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NO_INS]

    def get_number_of_logic_ins(self, block, graph) -> int:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NO_LOGIC_INS]

    def get_number_of_branch_ins(self, block, graph) -> int:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NO_BRANCH_INS]

    def get_str_consts(self, block, graph) -> List[str]:
        cache = self._get_correct_cache(graph)
        str_consts = cache[block][self.TYP_STR_CONST]
        if self._use_var_names:
            str_consts += cache[block][self.TYP_VAR_NAMES]
        if self._use_caller_names:
            str_consts += cache[block][self.TYP_FUNC_NAMES]

        return str_consts

    def get_num_consts(self, block, graph) -> List[int]:
        cache = self._get_correct_cache(graph)
        return cache[block][self.TYP_NUM_CONST] + cache[block][self.TYP_STACK_ADDRS]



from typing import List


class GenericStatement:
    def __init__(self, addr: int, operation: object, operands: List[object] = None):
        self.addr = addr
        self.op = operation
        self.operands = operands or []

    def __eq__(self, other):
        return isinstance(other, GenericStatement) and self.op == other.op and self.operands == self.operands

    def __hash__(self):
        _target = hash(self.addr) + hash(self.op)
        for operand in self.operands:
            _target += hash(operand)

        return _target

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f"<Statement {hex(self.addr)}: {self.op}({self.operands})"




# This is the CFGED (Control Flow GED) algorithm used in the paper:
# "Ahoy SAILR! There is No Need to DREAM of C: A Compiler-Aware Structuring Algorithm for Binary Decompilation"
# by Basque et al.
#
# The CFGED algorithm is intended to ONLY be used with Control Flow Graphs, that also have the following
# properties:
# 1. The graph is rooted, meaning it has a single start node and a single end node.
# 2. A mapping is provided that maps some nodes in g1 to some nodes in g2. This is used to guide the algorithm
#    to the correct regions to compute the GED on.
#
# This algorithm relies on the Abu-Aisheh GED algorithm, but it is modified to work with CFGs.

import itertools
from collections import defaultdict
from pathlib import Path
from typing import Dict, Union, Set, Optional, Tuple
import logging
import time
import os

import networkx as nx

from cfgutils.similarity import fast_is_isomorphic
from cfgutils.similarity.ged.abu_aisheh_ged import (
    ged_upperbound, graph_edit_distance_core_analysis, MAX_NODES_FOR_EXACT_GED, MIN_EXACT_TIME, MIN_UPPERBOUND_TIME,
)
from cfgutils.regions import GraphRegion, RegionIdentifier
from cfgutils.regions.utils import (
    expand_region_to_block_graph, expand_region_head_to_block, node_is_function_start, node_is_function_end
)
from cfgutils.data import GenericBlock
from cfgutils.file_formats import save_cfg_as_png
from cfgutils.transformations import to_supergraph


l = logging.getLogger(__name__)
_DEBUG = bool(os.getenv("DEBUG", False)) or False
if _DEBUG:
    l.setLevel(logging.DEBUG)

DEFAULT_LARGE_TIMEOUT = MIN_EXACT_TIME * 2

def toggle_debug():
    global _DEBUG
    _DEBUG = not _DEBUG
    l.setLevel(logging.DEBUG if _DEBUG else logging.INFO)


def addr_to_node_map(graph):
    # TODO: this is bugged, you can't list nodes by addr,
    # there needs to also be an index for nodes that share the same addr
    return {
        node.addr: node for node in graph.nodes
    }


def find_dst_mismatches(
    dec_edges, src_edges,
    dec_line_to_addr_map: Dict,
    src_addr_to_line_map: Dict,
):
    """
    In this algorithm, we always assume the decompiler is the graph we are transforming into the source graph.
    The basics of this idea is that after we've completed a region collapse, we want to find the edges that were
    not included in the score computation.

    Conceptually, every edge that _enters_ the region can can only enter at one place: the head. This means that when
    the region is collapsed, all those edges will remain going to that node in the right place (since both graphs
    pointed to the head).

    However, the outgoing edges from the region can come from any node in the region to the outside. This means that
    when we collapse the region we lose info. Here is an example of turning G1 into G2:
    G1:
    A -> B -> C
    |         ^
    ----------|

    G2:
    A -> B -> C
    |         ^
    ---> D ---|

    The region GED is 2. After the collapse, we have the following:
    G1:
    R1 -> C

    G2:
    R1 -> C

    The final CFGED score of this graph would be 2, but the real GED is 4.
    The edits needed for G1 -> G2 is: [del: (A, C), add: (D), add: (D, C), add: (A, D)]

    When we collapse the region, we lost information that (D, C) was the edge that was supposed to exist,
    not (A, C). Knowing that, would mean we need to del: (A,C) and then add: (D, C), worth 2 points.

    This leads to how to finally compute this:
    1. Collect all the edges leaving both graphs regions
    2. For each edge, find the corresponding edge in the other graph by their src node
    3. Remove then from the mismatch list
    4. Sum the number of remaining edges, that is the mismatch score
    """
    dec_edges = dec_edges["out"].copy()
    src_edges = src_edges["out"].copy()
    matches = set()
    already_matched_src_edge = set()
    for dec_src, dec_dst in dec_edges:
        dec_src_addrs = dec_line_to_addr_map.get(dec_src.addr, set())
        if not dec_src_addrs:
            continue

        src_src_addrs = set(
            itertools.chain.from_iterable([src_addr_to_line_map.get(addr, set()) for addr in dec_src_addrs]))
        if not src_src_addrs:
            continue

        for src_src, src_dst in src_edges:
            if (src_src, src_dst) in already_matched_src_edge:
                continue

            if src_src.addr in src_src_addrs:
                already_matched_src_edge.add((src_src, src_dst))
                matches.add((dec_src, dec_dst))
                matches.add((src_src, src_dst))
                break

    unmatched_dec_edges = list(filter(lambda x: x not in matches, dec_edges))
    unmatched_src_edges = list(filter(lambda x: x not in matches, src_edges))
    total_edits = len(unmatched_dec_edges) + len(unmatched_src_edges)

    return total_edits


def extra_points_for_func_end_edges(dec_r_cfg, src_r_cfg, dec_succ, src_succ):
    dec_is_end = node_is_function_end(dec_succ)
    src_is_end = node_is_function_end(src_succ)

    # in the event that both are either not the end or are the end
    # then we can depend on GED to handle it normally
    if not dec_is_end and not src_is_end:
        return 0
    elif dec_is_end and src_is_end:
        return 0

    dec_end_edges = list(dec_r_cfg.predecessors(dec_succ))
    src_end_edges = list(src_r_cfg.predecessors(src_succ))
    return len(dec_end_edges) + len(src_end_edges)


def destroy_old_region(cfg: nx.DiGraph, expanded_region_graph: nx.DiGraph, r_head: GenericBlock):
    extra_edges = defaultdict(list)
    r_nodes = list(expanded_region_graph.nodes)
    if r_head not in r_nodes:
        r_nodes.append(r_head)

    r_preds = list()
    r_succs = list()

    for r_node in r_nodes:
        for suc in cfg.successors(r_node):
            if suc not in r_nodes:
                r_succs.append(suc)
                extra_edges["out"].append((r_node, suc))
        for pred in cfg.predecessors(r_node):
            if pred not in r_nodes:
                r_preds.append(pred)
                extra_edges["in"].append((pred, r_node))

    cfg.remove_nodes_from(r_nodes)
    merged_node = r_head.merge_many_blocks(r_head.addr, r_nodes)
    cfg.add_node(merged_node, node=merged_node)
    for pred in r_preds:
        cfg.add_edge(pred, merged_node, src=pred, dst=merged_node)

    for suc in r_succs:
        cfg.add_edge(merged_node, suc, src=merged_node, dst=suc)

    return extra_edges


def expand_region_to_successor_block_graph(
    region: GraphRegion, full_graph: nx.DiGraph, include_preds=True, include_succs=True
) -> Tuple[Optional[nx.DiGraph], Optional[GenericBlock]]:
    # all regions should only every have 1 or None successors
    if region.successors and len(region.successors) > 1:
        return None, None

    #
    # Gather all the blocks inside the region
    #

    region_blocks = expand_region_to_blocks(region)

    #
    # Gather every block that has an edge into the region
    #

    pred_blocks = []
    if include_preds:
        for region_block in region_blocks:
            for pred in full_graph.predecessors(region_block):
                if pred not in region_blocks:
                    pred_blocks.append(pred)

    #
    # Gather every block that has an incoming edge from the region (successors)
    #

    # should end with a size of just 1
    successors = []
    if include_succs and region.successors and region.graph_with_successors:
        # assumption: the region has only one successor
        successor = list(region.successors)[0]
        if isinstance(successor, GenericBlock):
            successor_head = successor
        elif isinstance(successor, GraphRegion):
            successor_head = expand_region_head_to_block(successor)
        else:
            return None, None

        successors = [successor_head]
    successor_block = successors[0] if successors else None

    return nx.subgraph(full_graph, pred_blocks + region_blocks + successors), successor_block


def expand_region_to_blocks(_region: GraphRegion):
    all_nodes = list()
    for node in _region.graph.nodes:
        if isinstance(node, GenericBlock):
            all_nodes.append(node)
        elif isinstance(node, GraphRegion):
            all_nodes += expand_region_to_blocks(node)

    return all_nodes


def expand_region_to_block_graph(region: GraphRegion, graph: nx.DiGraph):
    region_blocks = expand_region_to_blocks(region)
    return nx.subgraph(graph, region_blocks)


def is_only_blocks(region: GraphRegion):
    for node in region.graph.nodes:
        if isinstance(node, GenericBlock):
            continue
        if isinstance(node, GraphRegion):
            return False

    return True


def find_containing_block_addrs(graph: nx.DiGraph, lines: Set):
    containing_addrs = set()
    line_has_container = set()
    graph_nodes = list(graph.nodes)
    for node in graph_nodes:
        for line in lines:
            if node.contains_addr(line):
                line_has_container.add(line)
                containing_addrs.add(node.addr)

    for line in lines:
        if line in line_has_container:
            continue

        closest_node = min(graph_nodes, key=lambda x: x.addr)
        for node in graph_nodes:
            if line >= node.addr >= closest_node.addr:
                closest_node = node

        containing_addrs.add(closest_node.addr)
    return containing_addrs


def find_matching_regions_with_lines(region: GraphRegion, lines: Set):
    if region.head.addr in lines:
        yield region

    for node in region.graph.nodes:
        if isinstance(node, GenericBlock):
            continue
        if isinstance(node, GraphRegion):
            yield from find_matching_regions_with_lines(node, lines)


def dfs_region_for_parent(region: GraphRegion, child: GenericBlock):
    for node in region.graph.nodes:
        if isinstance(node, GraphRegion):
            if node.head.addr == child.addr:
                yield region

            yield from dfs_region_for_parent(node, child)


def dfs_region_for_leafs(region: GraphRegion):
    has_block = False
    only_blocks = True
    for node in region.graph.nodes:
        if isinstance(node, GraphRegion):
            yield from dfs_region_for_leafs(node)
            only_blocks = False
        elif isinstance(node, GenericBlock):
            has_block = True

    if only_blocks and has_block:
        yield region


def find_some_tail_region(region: Union[GenericBlock, GraphRegion], node_blacklist, og_cfg: nx.DiGraph) -> Optional[
    Union[GraphRegion, GenericBlock]]:
    # sanity check
    if isinstance(region, GenericBlock):
        return region
    elif not isinstance(region, GraphRegion):
        return None

    leaf_regions = list(dfs_region_for_leafs(region))
    if not leaf_regions:
        return None

    # find a region we did not blacklist
    leaf_regions = sorted(leaf_regions, key=lambda x: x.head.addr, reverse=True)
    for leaf_region in leaf_regions:
        if leaf_region.head.addr not in node_blacklist:
            return leaf_region

    # if we are all out of non blacklisted regions, let's try to find a parent of each leaf
    # for leaf_region in leaf_regions:
    #    head, graph = leaf_region
    #    parents = list(dfs_region_for_parent(region, head))
    #    if not parents:
    #        continue

    #    parents = sorted(parents, key=lambda x: x.head.addr)
    #    for parent in parents:
    #        if parent.head.addr in node_blacklist:
    #            continue

    #        expanded_parent = expand_region_to_block_graph(parent, og_cfg)
    #        return parent.head, expanded_parent

    return None


def find_function_start_region(src_top_region: GraphRegion) -> Optional[GraphRegion]:
    """
    Finds a region, recursively, that has a function start node in it. A function start node
    is a node that has a statement that is a Nop with type FUNC_START.
    """
    if not isinstance(src_top_region, GraphRegion):
        return None

    for node in src_top_region.graph.nodes:
        if isinstance(node, GraphRegion):
            has_func_start = node_is_function_start(node.head)
            if has_func_start:
                return node

            matching_region = find_function_start_region(node)
            if matching_region is not None:
                return matching_region

    return None


def find_matching_source_region(
        dec_r_head, dec_r_cfg, src_regions, src_cfg, src_nodes, dec_line_to_addr_map, src_addr_to_line_map
) -> Tuple[Optional[GraphRegion], bool]:
    """
    :return (matching_src_region, redo_structuring)
    """

    # first do a sanity check to see if the dec head is a FUNCTION_START node, which can speedup analysis
    dec_head_is_func_start = node_is_function_start(dec_r_head)
    if dec_head_is_func_start:
        best_match = find_function_start_region(src_regions)
        if best_match is not None:
            return best_match, True

    # map the decompilation line to an address (reported by the decompiler
    addrs = dec_line_to_addr_map.get(dec_r_head.addr, None)
    if not addrs:
        # quick check if you can find the line in the addr map in a proximity
        up_addrs = dec_line_to_addr_map.get(dec_r_head.addr - 1, None)
        down_addrs = dec_line_to_addr_map.get(dec_r_head.addr + 1, None)
        if not up_addrs and not down_addrs:
            l.debug(f"Unable to find any line-addr map for region head {dec_r_head.addr}! Skipping...")
            return None, True

        addrs = up_addrs if up_addrs else down_addrs

    all_lines = set(
        itertools.chain.from_iterable(
            [src_addr_to_line_map[addr] for addr in addrs if addr in src_addr_to_line_map]
        )
    )
    lines_in_src = set(filter(lambda x: x in src_nodes, all_lines))

    # We were unable to find a real region start, probably because a node got consumed
    if not lines_in_src:
        lines_in_src = find_containing_block_addrs(src_cfg, all_lines)
        if not lines_in_src and addrs:
            # If we still have nothing, walk backwards!
            for i in range(1, 0x14):
                all_lines = set(
                    itertools.chain.from_iterable(
                        [src_addr_to_line_map[addr - i] for addr in addrs if addr - i in src_addr_to_line_map]
                    )
                )
                lines_in_src = set(filter(lambda x: x in src_nodes, all_lines))
                if lines_in_src:
                    break

    lines_in_src = sorted(lines_in_src)
    matching_src_regions = list(find_matching_regions_with_lines(src_regions, lines_in_src))
    if not matching_src_regions:
        l.debug(f"Unable to find a pairing region for {dec_r_head.addr}: no src addrs found")
        return None, False

    #
    # try to find the base matching region
    #

    # gather sizes
    matches_by_size = {}
    lowest_size = 10000
    for src_region in matching_src_regions:
        src_block_region = expand_region_to_block_graph(src_region, src_cfg)
        src_r_head = src_region.head
        size_diff = abs(len(src_block_region.nodes) - len(dec_r_cfg.nodes))
        matches_by_size[src_r_head] = (src_region, size_diff)
        if size_diff < lowest_size:
            lowest_size = size_diff

    # filter out the ones that are too big
    matching_src_regions = list(filter(lambda x: x[1][1] <= lowest_size, matches_by_size.items()))
    if not matching_src_regions:
        l.debug(f"Unable to find a pairing region for {dec_r_head.addr}")
        return None, False

    # if we have more than one, tie break with statement size
    best_match = None
    if len(matching_src_regions) > 1:
        smallest_stmt_size = 10000
        for (head, (region, size)) in matching_src_regions:
            if not isinstance(head, GenericBlock):
                continue

            has_merged_region = head.is_merged_node
            if len(head.statements) < smallest_stmt_size:
                smallest_stmt_size = len(head.statements)
                best_match = region

            if has_merged_region:
                best_match = region
                break

    if best_match is None:
        best_match = matching_src_regions[0][1][0]

    # if we matched our decompilation to source and the head is a function start
    # than we likly made a mistake in approximation of matches
    # if not dec_head_is_func_start and node_is_function_start(best_match.head):
    #    return None, False

    return best_match, True


def fast_switch_region_ged(expanded_dec_r_graph, expanded_src_r_graph):
    """
    The intuition behind this is that in the event we know we are matching two switches across two
    graphs, we know all the first-level children of the graph going to be the same and their parent is the same.
    Thus, we can eliminate all (first_parent, childN) edges. Now all we have left is:
    1. Diff in # of cases between both switches (nodes)
    2. Diff in case successors (edges)

    In most cases, this will be a very small number of edits. We can simply approximate it by counting # of
    node and edge differences since those will make up all of the difference in the exit of the region.
    """

    return abs(len(expanded_dec_r_graph.nodes) - len(expanded_src_r_graph.nodes)) + abs(
        len(expanded_dec_r_graph.edges) - len(expanded_src_r_graph.edges))


def is_switch_region(expanded_region_graph: nx.DiGraph):
    if expanded_region_graph is None:
        return False

    region_heads = [node for node in expanded_region_graph.nodes if expanded_region_graph.in_degree(node) == 0]
    if len(region_heads) != 1:
        return False

    region_head = region_heads[0]
    return expanded_region_graph.out_degree(region_head) > 3


def cfg_edit_distance(
    dec_cfg: nx.DiGraph,
    src_cfg: nx.DiGraph,
    dec_line_to_addr_map: Dict,
    src_addr_to_line_map: Dict,
    max_region_collapse=200,
    max_region_estimates=3,
    check_upperbound_approx=False,
    timeout=None,
):
    """
    src_addr_to_line_map[address] = set(line1, line2, ...)
    dec_line_to_addr_map[line_num] = set(addr1, addr2, ...)

    :param src_cfg:
    :param dec_cfg:
    :param src_addr_to_line_map:
    :param dec_line_to_addr_map:
    :param max_region_collapse:
    :return:
    """
    src_cfg, dec_cfg = nx.DiGraph(src_cfg), nx.DiGraph(dec_cfg)

    if fast_is_isomorphic(dec_cfg, src_cfg):
        l.debug(f"Graphs are isomorphic! Returning 0...")
        return 0

    cfged_score = 0
    curr_region_collapse = 0
    curr_region_estimates = 0

    # will always be set on the first iteration since redo_structuring is True
    src_regions, dec_regions = None, None
    src_nodes, dec_nodes = {}, {}
    redo_structuring = True

    unable_to_approx = False
    region_blacklist = set()

    # if len(src_cfg.nodes) <= MAX_EXACT_GED_SIZE and len(dec_cfg.nodes) <= MAX_EXACT_GED_SIZE:
    #    l.debug(f"Graph small enough for exact! Running it...")
    #    return graph_edit_distance(dec_cfg, src_cfg, with_timeout=10, max_on_timeout=False)

    # the max possible GED score for any two graphs
    max_cfged_score = len(src_cfg.nodes) + len(dec_cfg.nodes) + len(src_cfg.edges) + len(dec_cfg.edges)
    upperbound = max_cfged_score
    if check_upperbound_approx:
        approx = ged_upperbound(dec_cfg, src_cfg, with_timeout=MIN_UPPERBOUND_TIME)
        if approx is not None:
            upperbound = approx

    start_time = time.time()
    while True:
        curr_time = time.time()
        if unable_to_approx or curr_region_collapse >= max_region_collapse:
            if unable_to_approx:
                l.debug(f"Unable to approximate the rest of the graph. Exiting early with max score!")
            else:
                l.debug(f"Max region collapse limit hit. Exiting early with max score!")
            return cfged_score + len(src_cfg.nodes) + len(dec_cfg.nodes) + len(src_cfg.edges) + len(dec_cfg.edges)

        l.debug(f"Running region collapse round {curr_region_collapse}...")
        if redo_structuring:
            # src_cfg, dec_cfg = to_jil_supergraph(src_cfg), to_jil_supergraph(dec_cfg)
            src_nodes, dec_nodes = addr_to_node_map(src_cfg), addr_to_node_map(dec_cfg)
            l.debug(f"Decompiler nodes: {len(dec_cfg.nodes)}")
            l.debug(f"Source nodes: {len(src_cfg.nodes)}")

            # compute the regions
            src_regions, dec_regions = RegionIdentifier(src_cfg).region, RegionIdentifier(dec_cfg).region

        if _DEBUG:
            save_cfg_as_png(src_cfg, Path(f"./src_cfg_{curr_region_collapse}.png"))
            save_cfg_as_png(dec_cfg, Path(f"./dec_cfg_{curr_region_collapse}.png"))

        #
        # Find a tail region pair across both graphs
        #

        # Now that you have regions we want to iteratively enter the lower region we can find on one graph
        # then find that same region on the other graph, then do a graph edit distance. In this case
        # we start with the src cfg
        dec_region = find_some_tail_region(dec_regions, region_blacklist, dec_cfg)
        if dec_region is None:
            dec_r_head, dec_r_cfg = None, None
        elif isinstance(dec_region, GenericBlock):
            dec_r_head, dec_r_cfg = dec_region, nx.DiGraph()
            dec_r_cfg.add_node(dec_region)
        else:
            dec_r_head, dec_r_cfg = dec_region.head, dec_region.graph

        if (dec_r_head is None) or (timeout is not None and curr_time - start_time > timeout):
            big_timeout = DEFAULT_LARGE_TIMEOUT
            if dec_r_head is None:
                l.debug(f"We are unable to match anymore small region heads, which means we must now approximate the rest.")
                if timeout is not None:
                    time_left = timeout - (curr_time - start_time)
                    if time_left > DEFAULT_LARGE_TIMEOUT:
                        big_timeout = time_left
            else:
                l.debug(f"Timeout hit! Approximating the rest of the graph...")

            if _DEBUG:
                save_cfg_as_png(src_cfg, Path(f"./src_cfg_{curr_region_collapse}.png"))
                save_cfg_as_png(dec_cfg, Path(f"./dec_cfg_{curr_region_collapse}.png"))

            score = graph_edit_distance_core_analysis(dec_cfg, src_cfg, with_timeout=big_timeout)
            if score is None:
                unable_to_approx = True
                continue

            cfged_score += score
            break

        best_match, redo_structuring = find_matching_source_region(
            dec_r_head, dec_r_cfg, src_regions, src_cfg, src_nodes, dec_line_to_addr_map, src_addr_to_line_map
        )
        if redo_structuring:
            curr_region_collapse += 1
        if best_match is None:
            region_blacklist.add(dec_r_head.addr)
            continue

        l.debug(f"Collapsing (Dec, Src) region pair: {(dec_r_head, best_match.head)}")

        #
        # compute GED of the expanded-matched tail regions
        #

        src_r_cfg = expand_region_to_block_graph(best_match, src_cfg)
        expanded_src_graph, src_successor = expand_region_to_successor_block_graph(best_match, src_cfg)
        expanded_dec_graph, dec_successor = expand_region_to_successor_block_graph(dec_region, dec_cfg)

        bad_successor_case = expanded_src_graph is None or expanded_dec_graph is None
        if not bad_successor_case:
            dec_size, src_size = len(expanded_dec_graph.nodes), len(expanded_src_graph.nodes)
        else:
            dec_size, src_size = len(dec_r_cfg.nodes), len(src_r_cfg.nodes)

        too_large_for_exact = dec_size > MAX_NODES_FOR_EXACT_GED or src_size > MAX_NODES_FOR_EXACT_GED
        # a heuristic to see if the size difference is too large to approximate, which will cause a skip
        if too_large_for_exact:
            size_diff = abs(dec_size - src_size)
            if size_diff > dec_size * 2 or size_diff > src_size * 2:
                l.debug(
                    f"Difference in regions size too large to approximate (Dec, Src): ({dec_size}, {src_size}), skipping for now...")
                region_blacklist.add(dec_r_head.addr)
                redo_structuring = False
                continue
            else:
                l.debug("Region size is too large for exact. This difference will be an approximation.")
                curr_region_estimates += 1

        # In most graph expasions of a region, the region should have a single successor and be moderately sized.
        # In the event this is not true, we can generate highly innacurate GED scores because we rely on GED
        # approximation of this larger/broken region. To mitigate this, we can special case scenarios which cause
        # the most problems
        #
        # 1. The region is a switch statement, which means its region is huge but actually quite simple
        switch_region_case = is_switch_region(expanded_src_graph) and is_switch_region(
            expanded_dec_graph) and too_large_for_exact
        # 2. The regions has multiple successors (bad_successor_case)

        if switch_region_case:
            l.debug("Special Case Detected: Switch Region. Using Fast Switch GED...")
            distance = fast_switch_region_ged(expanded_dec_graph, expanded_src_graph)
        elif bad_successor_case:
            l.debug("Special Case Detected: Multi-Successor Region. Using edge-diffs for approx...")
            distance = graph_edit_distance_core_analysis(dec_r_cfg, src_r_cfg, with_timeout=MIN_EXACT_TIME)
        else:
            distance = graph_edit_distance_core_analysis(expanded_dec_graph, expanded_src_graph, with_timeout=MIN_EXACT_TIME)

        if distance is None:
            l.debug(f"Unable to compute the GED of the region, skipping...")
            region_blacklist.add(dec_r_head.addr)
            redo_structuring = False
            continue

        cfged_score += distance
        l.debug(f"Region distance: {distance}")

        #
        # cleanup, destroy measured nodes, and restart loop
        #

        extra_src_edges = destroy_old_region(
            src_cfg,
            expand_region_to_block_graph(best_match, src_cfg),
            expand_region_head_to_block(best_match)
        )
        extra_dec_edges = destroy_old_region(
            dec_cfg,
            expand_region_to_block_graph(dec_region, dec_cfg) if not isinstance(dec_region, GenericBlock) else dec_region,
            expand_region_head_to_block(dec_region) if not isinstance(dec_region, GenericBlock) else dec_region
        )

        # extra edge diffs are only needed in the event we can't use a normally expanded region graph
        if bad_successor_case:
            edge_diff = find_dst_mismatches(extra_dec_edges, extra_src_edges, dec_line_to_addr_map,
                                            src_addr_to_line_map)
            if edge_diff:
                l.debug(f"In/Out Edge Diff: {edge_diff}")
                cfged_score += edge_diff

        # if we have only a single node left in either graph, we can just add the remaining edges and nodes to
        # the score (since there is only one way to construct the graph)
        if len(dec_cfg.nodes) <= 1:
            cfged_score += (len(src_cfg.nodes) + len(src_cfg.edges)) - len(dec_cfg.nodes)
            break
        elif len(src_cfg.nodes) <= 1:
            cfged_score += (len(dec_cfg.nodes) + len(dec_cfg.edges)) - len(src_cfg.nodes)
            break

        # reset blacklist and structuring check
        region_blacklist = set()
        redo_structuring = True

    l.debug(f"Final CFGED Score: {cfged_score}")
    # handle error cases.
    # this only will make sure scores from CFGED are always right, but these cases should all
    # be investigated and fixed if encountered.
    if cfged_score > max_cfged_score:
        l.critical(f"CFGED score of {cfged_score} is somehow larger than the max possible score! DEBUG THIS!")
        return max_cfged_score
    elif check_upperbound_approx and cfged_score > upperbound:
        l.debug(f"CFGED score {cfged_score} is larger than upperbound {upperbound}, returning upperbound")
        return upperbound
    else:
        return cfged_score



from typing import List, Tuple

import networkx as nx

from .generic_block import GenericBlock
from .generic_statement import GenericStatement


def numbered_edges_to_block_graph(numbered_edges: List[Tuple[int, int]]) -> nx.DiGraph:
    """
    Node numbering should start at 1. If a number is in the form of a float, like 1.1, then the number on
    the right of the decimal will be treated as the idx, which is a unique identifier. Please use small
    numbers for the block addresses.
    """

    # find max block number to create a block dictionary
    block_numbers = set()
    float_edges = []
    for src, dst in numbered_edges:
        block_numbers.add(src)
        block_numbers.add(dst)
        if type(src) is float or type(dst) is float:
            float_edges.append((src, dst))

    max_number = int(max(block_numbers))
    # None blocks added to make indexing for the right block addr easier
    blocks = [None] + [GenericBlock(i) for i in range(1, max_number+1)] + [None]

    # do all normal edges
    block_edges = [
        (blocks[in_e], blocks[out_e]) for (in_e, out_e) in numbered_edges
        if not type(in_e) is float and not type(out_e) is float
    ]
    # do all float edges (extra data)
    if float_edges:
        float_blocks = {}
        for edge in float_edges:
            for node in edge:
                if type(node) is float:
                    float_str = str(node)
                    if float_str not in float_blocks:
                        idx = int(float_str.split(".")[-1])
                        float_blocks[float_str] = GenericBlock(int(node), idx=idx)

        for src, dst in float_edges:
            src_blk = float_blocks[str(src)] if type(src) is float else blocks[src]
            dst_blk = float_blocks[str(dst)] if type(dst) is float else blocks[dst]
            block_edges.append((src_blk, dst_blk))

    graph = nx.DiGraph()
    graph.add_edges_from(block_edges)

    # find start and ends and update their attributes
    starts = [n for n in graph.nodes if graph.in_degree(n) == 0]
    ends = [n for n in graph.nodes if graph.out_degree(n) == 0]
    for node in starts:
        node.is_entrypoint = True
    for node in ends:
        node.is_exitpoint = True

    # update the node attr of every node in nx to be itself
    for node in graph.nodes:
        graph.nodes[node]["node"] = node

    # update the edge attr of every edge in nx to be itself
    for edge in graph.edges:
        graph.edges[edge]["src"] = edge[0]
        graph.edges[edge]["dst"] = edge[1]

    return graph




import sys
import unittest

import networkx as nx

from cfgutils.data import GenericBlock, GenericStatement
from cfgutils.hashing.function_sim_search import FunctionSimHasher


class TestHashing(unittest.TestCase):
    def test_fss_hashing(self):
        """
        Tests variations of the exact same graph, but with different edges excluded.
        The code the graph represents vaguely looks like this:

        x = 0x13
        if x >= 0x1337 {
            y = 0x14;
        }
        else {
            y = 0x15;
        }
        return;
        """
        blocks = [
            # blk 1
            GenericBlock(0x1, statements=[
                GenericStatement(0x2, "assign", ["x", 0x13]),
                GenericStatement(0x3, "cond_jmp", ["x", "gte", 0x1337])
            ]),
            # blk 2
            GenericBlock(0x4, statements=[
                GenericStatement(0x5, "assign", ["y", 0x14]),
            ]),
            # blk 3
            GenericBlock(0x6, statements=[
                GenericStatement(0x7, "assign", ["y", 0x15]),
            ]),
            # blk 4
            GenericBlock(0x8, statements=[
                GenericStatement(0x9, "ret", []),
            ]),
        ]
        b1, b2, b3, b4 = blocks
        # full diamond graph
        g1 = nx.DiGraph()
        g1.add_edges_from(
            [(b1, b2), (b1, b3), (b2, b4), (b3, b4)]
        )

        # partial diamond
        g2 = nx.DiGraph()
        g2.add_edges_from(
            [(b1, b2), (b1, b3), (b2, b4)]
        )

        # triangle
        g3 = nx.DiGraph()
        g3.add_edges_from(
            [(b1, b2), (b1, b3)]
        )

        # line
        g4 = nx.DiGraph()
        g4.add_edges_from(
            [(b1, b2)]
        )

        # dot!
        g5 = nx.DiGraph()
        g5.add_node(b1)

        hasher = FunctionSimHasher()
        h1 = hasher.CalculateFunctionSimHash(g1)
        h2 = hasher.CalculateFunctionSimHash(g2)
        h3 = hasher.CalculateFunctionSimHash(g3)
        h4 = hasher.CalculateFunctionSimHash(g4)
        h5 = hasher.CalculateFunctionSimHash(g5)

        # each distance should be less than the last
        d11 = FunctionSimHasher.hash_distance(h1, h1)
        d12 = FunctionSimHasher.hash_distance(h1, h2)
        d13 = FunctionSimHasher.hash_distance(h1, h3)
        d14 = FunctionSimHasher.hash_distance(h1, h4)
        d15 = FunctionSimHasher.hash_distance(h1, h5)

        assert d11 == 0
        assert d11 <= d12
        assert d12 <= d13
        # TODO: find out why things fall apart when the graph is too small!
        #assert d13 <= d14
        #assert d14 <= d15


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



from cfgutils.sorting import cfg_root_node

INVALID_CHOICE_PENALTY = 100000

#
# Helpers
#


def collect_graph_roots(g1, g2):
    # first, depend on the function start node
    g1_start, g2_start = cfg_root_node(g1), cfg_root_node(g2)
    if g1_start is not None and g2_start is not None:
        roots = (g1_start, g2_start,)
    else:
        roots = None

    # second attempt, use predecessors
    if roots is None:
        g1_starts = list(node for node in g1.nodes if len(list(g1.predecessors(node))) == 0)
        g2_starts = list(node for node in g2.nodes if len(list(g2.predecessors(node))) == 0)
        if len(g1_starts) == 1 == len(g2_starts):
            roots = (g1_starts[0], g2_starts[0],)

    return roots





from typing import Optional, Any

from ailment import BinaryOp, UnaryOp, Assignment
from ailment.block_walker import AILBlockWalkerBase
from ailment.block import Block
from ailment.statement import Call, ConditionalJump, Statement, Store
from ailment.expression import Const, StackBaseOffset, BasePointerOffset, Load

from .prettyify_ail import string_at_addr


class AILBlockFeatureExtractor(AILBlockWalkerBase):
    """
    Extracts features from AIL blocks based on discoveRE features
    """
    LOGIC_INS_TYPS = {
        "LogicalAnd", "LogicalOr", "CmpF", "CmpEQ", "CmpNE", "CmpLT", "CmpLE", "CmpGT", "CmpGE", "CmpLTs",
        "CmpLEs", "CmpGTs", "CmpGEs"
    }
    ARITH_INS_TYPS = {
        "Add", "AddF", "Sub", "SubF", "Mul", "MulF", "Div", "DivF", "DivMo", "Mod", "Xor", "And",
        "Or", "Shl", "Shr", "Sar", "Ror", "Rol", "Not"
    }

    def __init__(self, project=None, project_cfg=None, call_name_fallback=None):
        self._project = project
        self._project_cfg = project_cfg
        self._call_name_fallback_addrs = call_name_fallback or {}

        # features
        self.arith_ins = []
        self.calls = []
        self.ins = []
        self.logic_ins = []
        self.branch_ins = []
        self.str_consts = []
        self.num_consts = []
        self.stack_addrs = []
        self.var_names = []
        self.call_names = []

        super().__init__()
        self.expr_handlers[StackBaseOffset] = self._handle_StackBaseOffset

    def _handle_stmt(self, stmt_idx: int, stmt: Statement, block: Optional[Block]) -> Any:
        self.ins.append(stmt)
        return super()._handle_stmt(stmt_idx, stmt, block)

    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
        if expr.op in self.ARITH_INS_TYPS:
            self.arith_ins.append(expr)
        elif expr.op in self.LOGIC_INS_TYPS:
            self.logic_ins.append(expr)
        return super()._handle_BinaryOp(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_UnaryOp(self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement, block: Optional[Block]):
        if expr.op in self.ARITH_INS_TYPS:
            self.logic_ins.append(expr)
        return super()._handle_UnaryOp(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_ConditionalJump(self, stmt_idx: int, stmt: ConditionalJump, block: Optional[Block]):
        self.branch_ins.append(stmt)
        return super()._handle_ConditionalJump(stmt_idx, stmt, block)

    def _handle_CallExpr(self, expr_idx: int, expr: "Call", stmt_idx: int, stmt, block: Optional["Block"]):
        self.calls.append(expr)
        func_name = self._resolve_call_name(expr)
        if func_name is not None:
            self.call_names.append(func_name)
        super()._handle_CallExpr(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Call(self, stmt_idx: int, stmt: "Call", block: Optional["Block"]):
        self.calls.append(stmt)
        func_name = self._resolve_call_name(stmt)
        if func_name is not None:
            self.call_names.append(func_name)
        super()._handle_Call(stmt_idx, stmt, block)

    def _handle_Const(self, expr_idx: int, expr: "Const", stmt_idx: int, stmt: Statement, block: Optional[Block]):
        if self._project is not None and self._project_cfg is not None:
            str_val = string_at_addr(self._project_cfg, expr.value, self._project, max_size=200)
            if str_val is not None:
                self.str_consts.append(str_val)
            else:
                self.num_consts.append(expr.value)
        else:
            self.num_consts.append(expr.value)

        return super()._handle_Const(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_StackBaseOffset(self, expr_idx: int, expr: "StackBaseOffset", stmt_idx: int, stmt: Statement, block: Optional[Block]):
        self.stack_addrs.append(expr.offset)
        return None

    def _handle_Store(self, stmt_idx: int, stmt: Store, block: Optional[Block]):
        if stmt.variable is not None and stmt.variable.name is not None:
            self.var_names.append(stmt.variable.name)

        return super()._handle_Store(stmt_idx, stmt, block)

    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement, block: Optional[Block]):
        if hasattr(expr, "variable") and expr.variable is not None and expr.variable.name is not None:
            self.var_names.append(expr.variable.name)

        return super()._handle_Load(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Optional[Block]):
        if hasattr(stmt.dst, "variable") and stmt.dst.variable is not None and stmt.dst.variable.name is not None:
            self.var_names.append(stmt.dst.variable.name)

        if hasattr(stmt.src, "variable") and stmt.src.variable is not None and stmt.src.variable.name is not None:
            self.var_names.append(stmt.src.variable.name)

        return super()._handle_Assignment(stmt_idx, stmt, block)

    #
    # utils
    #

    def _resolve_call_name(self, call_expr):
        func_name = None
        if isinstance(call_expr.target, Const):
            func_addr = call_expr.target.value
            if self._project is not None and func_addr in self._project.kb.functions and not self._project.kb.functions[func_addr].name.startswith("sub_"):
                func_name = self._project.kb.functions[func_addr].name
            elif func_addr in self._call_name_fallback_addrs:
                func_name = self._call_name_fallback_addrs[func_addr]
        if isinstance(call_expr.target, Load) and isinstance(call_expr.target.addr, StackBaseOffset):
            # convert to lookup format
            k = f"s_{hex(call_expr.target.addr.offset)}"
            func_name = self._call_name_fallback_addrs.get(k, None)

        return func_name



import sys
import unittest
from pathlib import Path

import networkx as nx

ANGR_AVAILABLE = True
try:
    import angr
except ImportError:
    ANGR_AVAILABLE = False

if ANGR_AVAILABLE:
    from cfgutils.angr_utils.ail_graph_conv import binary_to_ail_cfgs
    from cfgutils.angr_utils.block_matcher import AILBlockMatcher
    from angr.analyses.decompiler.utils import find_block_by_addr

TEST_FILES = Path(__file__).parent / "data"


class TestAngrCFGTools(unittest.TestCase):
    def setUp(self):
        if not ANGR_AVAILABLE:
            self.skipTest("angr is not available")

    def test_cfg_attrs(self):
        cfgs = binary_to_ail_cfgs(
            TEST_FILES / "fmt_O0_noinline.o",
            functions=["fmt"],
            structuring_opts=False,
            make_generic=True
        )
        cfg: nx.DiGraph = cfgs["fmt"]
        for node, attr in cfg.nodes(data=True):
            assert attr is not None
            original_nodes = attr.get("original_nodes", {})
            assert original_nodes

    # TODO: re-enable this test later
    def DISABLE_test_block_matcher(self):
        cfgs0, p0 = binary_to_ail_cfgs(
            TEST_FILES / "fmt_O0_noinline.o",
            functions=["main"],
            structuring_opts=False,
            return_project=True
        )
        cfgs2, p2 = binary_to_ail_cfgs(
            TEST_FILES / "fmt_O2_noinline.o",
            functions=["main"],
            structuring_opts=False,
            return_project=True
        )
        main_o0 = cfgs0["main"]
        main_o2 = cfgs2["main"]

        assert main_o0 is not None
        assert main_o2 is not None

        matcher = AILBlockMatcher(main_o0, main_o2, proj1=p0, proj2=p2)
        mappings = matcher.mapping

        # TODO: add more assertions
        o0_blk = find_block_by_addr(main_o0, 0x400862)
        o2_blk = find_block_by_addr(main_o2, 0x40ce2b)
        assert mappings[o0_blk] == o2_blk


if __name__ == "__main__":
    unittest.main(argv=sys.argv)



import pickle
from pathlib import Path
import logging
from typing import Union, Dict, Tuple

import ailment
import networkx as nx
import angr
from angr.analyses.decompiler.optimization_passes import DUPLICATING_OPTS, CONDENSING_OPTS

from .prettyify_ail import stmt_to_pretty_text
from cfgutils.data.generic_block import GenericBlock

_l = logging.getLogger(__name__)


def _set_decompiler_option(decompiler_options_strs: tuple[str, object]) -> list[tuple]:
    decompiler_options = []
    if not decompiler_options_strs:
        return decompiler_options

    for param, value in decompiler_options_strs:
        for option in angr.analyses.decompiler.decompilation_options.options:
            if param == option.param:
                decompiler_options.append((option, value))

    return decompiler_options


def binary_to_ail_cfgs(
    binary_path: Path, functions=None, make_generic=False, structuring_opts=True, supergraph=True,
    return_project=False, decompiler_options=None
) -> Union[Dict[str, nx.DiGraph], Tuple[Dict[str, nx.DiGraph], angr.Project]]:
    """
    A simple wrapper around the angr decompiler to simply use the defaults and return the AIL CFGs which
    are present at the end of the decompilation process. Using make_generic, you can convert the AIL CFGs
    into GenericBlock CFGs for use with CFGUtils functions. You can also disable some of the structuring
    optimizations if you want to see the raw CFGs.

    :param binary_path: Path to the binary to decompile
    :param functions: List of function addresses to decompile. If None, all functions will be decompiled
    :param make_generic: Convert the AIL CFGs to GenericBlock CFGs
    :param structuring_opts: Enable or disable structuring optimizations
    :param supergraph: Convert the AIL CFGs to supergraphs
    :param return_project: Return the angr Project object as well
    :param decompiler_options: A list of tuples containing the decompiler options to set. Each tuple should be
        in the format (option_name, option_value). See angr documentation for available options.
    """
    binary_path = Path(binary_path).absolute()
    if not binary_path.exists():
        raise FileNotFoundError(f"{binary_path} does not exist")

    proj = angr.Project(binary_path, auto_load_libs=False)
    func_starts = None
    if functions is not None and all(isinstance(func, int) for func in functions):
        func_starts = functions

    cfg = proj.analyses.CFG(
        show_progressbar=False, normalize=True, data_references=True, resolve_indirect_jumps=True,
        function_starts=func_starts
    )
    try:
        proj.analyses.CompleteCallingConventions(cfg=cfg, recover_variables=True, analyze_callsites=True)
        cc_failed = False
    except Exception:
        cc_failed = True

    if cc_failed:
        _l.warning(f"CallingConvention Analysis failed on {binary_path}. Trying again without variable recovery...")
        try:
            # try it again without variable recovery
            proj.analyses.CompleteCallingConventions(cfg=cfg, recover_variables=False)
            cc_failed = False
        except Exception:
            pass

    if cc_failed:
        _l.critical(f"All attempts to run CallingConvention Analysis failed on {binary_path}.")

    # clean up function names
    functions = functions or cfg.functions
    for func in list(cfg.functions.values()):
        if "." in func.name:
            func.name = func.name[:func.name.index(".")]

    # some optimizations can drastically change the structure of the CFG, so we should disable some if wanted
    all_optimizations = angr.analyses.decompiler.optimization_passes.get_optimization_passes(
        proj.arch, "linux"
    )
    if not structuring_opts:
        all_optimizations = [opt for opt in all_optimizations if opt not in DUPLICATING_OPTS + CONDENSING_OPTS]

    # generate a cfg for each function
    ail_cfgs = {}
    for func_idx, func_addr in enumerate(functions):
        try:
            f = cfg.functions[func_addr]
        except Exception:
            _l.critical(f"Function at {func_addr} not found in CFG")
            continue

        if f is None or f.is_plt:
            continue

        # for this function you don't actually need the linear decompilation, but we run through the entire
        # decompilation process to assure every optimization is run that would be done on a normal Clinic graph
        dec = proj.analyses.Decompiler(
            f, cfg=cfg, kb=cfg.kb, optimization_passes=all_optimizations, generate_code=False,
            options=_set_decompiler_option(decompiler_options)
        )
        dec.ail_graph.name = str(f.name)
        ail_cfgs[str(f.name)] = dec.ail_graph if not supergraph else to_ail_supergraph(dec.ail_graph)

    if make_generic:
        cfgs = [ail_cfg_to_generic(cfg, proj) for name, cfg in ail_cfgs.items()]
    else:
        cfgs = list(ail_cfgs.values())

    named_cfgs = {cfg.name or str(i): cfg for i, cfg in enumerate(cfgs)}
    if return_project:
        return named_cfgs, proj
    else:
        return named_cfgs


def ail_cfg_to_generic(cfg: nx.DiGraph, project=None):
    new_cfg = nx.DiGraph()
    new_cfg.name = cfg.name
    proj_cfg = project.kb.cfgs.get_most_accurate() if project is not None else None
    node_map = {}
    for node, attr in cfg.nodes(data=True):
        new_node = GenericBlock(node.addr, idx=node.idx if hasattr(node, "idx") else None)
        for stmt in node.statements:
            str_stmt = stmt_to_pretty_text(stmt, project, proj_cfg) if project is not None else str(stmt)
            new_node.statements.append(str_stmt)

        attr = attr or {}
        new_attr = attr.copy()
        if "node" in new_attr:
            del new_attr["node"]

        new_cfg.add_node(new_node, node=new_node, **new_attr)
        node_map[node] = new_node

    for src, dst in cfg.edges:
        new_src = node_map[src]
        new_dst = node_map[dst]
        new_cfg.add_edge(new_src, new_dst, src=new_src, dst=new_dst)

    return new_cfg


def binary_to_generic_cfgs(binary_path: Path, functions=None):
    return binary_to_ail_cfgs(binary_path, functions, make_generic=True)


def ail_pickle_to_cfg(pickle_path: Path):
    """
    This function converts a pickle files, containing an AIL CFG, into a GenericBlock CFG for use with
    CFGUtils functions. Note: this pickled AIL CFG is a little special. All the nodes in the CFG are just
    the addresses of the blocks (not real AIL nodes). The 'label' attribute of each node is a list of strings,
    which is simply each AIL stmt stringed.
    """

    with open(pickle_path, "rb") as fp:
        ail_str_cfg: nx.DiGraph = pickle.load(fp)

    new_cfg = nx.DiGraph()
    node_map = {}
    for node in ail_str_cfg.nodes:
        new_node = GenericBlock(int(node), statements=ail_str_cfg.nodes[node]["label"])
        new_cfg.add_node(new_node, node=new_node)
        node_map[node] = new_node

    for src, dst in ail_str_cfg.edges:
        new_src = node_map[src]
        new_dst = node_map[dst]
        new_cfg.add_edge(new_src, new_dst, src=new_src, dst=new_dst)

    return new_cfg


#
# TODO: all of the following functions need to be removed when  is fixed
#   this code was added as a hotfix to make the AIL CFGs supergraphs
#

def _merge_ail_nodes(graph, node_a: ailment.Block, node_b: ailment.Block) -> ailment.Block:
    in_edges = list(graph.in_edges(node_a, data=True))
    out_edges = list(graph.out_edges(node_b, data=True))

    a_ogs = graph.nodes[node_a].get("original_nodes", set())
    b_ogs = graph.nodes[node_b].get("original_nodes", set())
    new_node = node_a.copy() if node_a.addr <= node_b.addr else node_b.copy()
    old_node = node_b if new_node == node_a else node_a
    # remove jumps in the middle of nodes when merging
    if new_node.statements and isinstance(new_node.statements[-1], ailment.Stmt.Jump):
        new_node.statements = new_node.statements[:-1]
    new_node.statements += old_node.statements
    new_node.original_size += old_node.original_size

    graph.remove_node(node_a)
    graph.remove_node(node_b)

    if new_node is not None:
        graph.add_node(new_node, original_nodes=a_ogs.union(b_ogs))
        for src, _, data in in_edges:
            if src is node_b:
                src = new_node
            graph.add_edge(src, new_node, **data)

        for _, dst, data in out_edges:
            if dst is node_a:
                dst = new_node
            graph.add_edge(new_node, dst, **data)

    return new_node


def to_ail_supergraph(transition_graph: nx.DiGraph) -> nx.DiGraph:
    """
    Takes an AIL graph and converts it into a AIL graph that treats calls and redundant jumps
    as parts of a bigger block instead of transitions. Calls to returning functions do not terminate basic blocks.

    Based on region_identifier super_graph

    :return: A converted super transition graph
    """
    # make a copy of the graph
    transition_graph = nx.DiGraph(transition_graph)
    nx.set_node_attributes(transition_graph, {node: {node} for node in transition_graph.nodes}, "original_nodes")

    while True:
        for src, dst, data in transition_graph.edges(data=True):
            type_ = data.get("type", None)

            if len(list(transition_graph.successors(src))) == 1 and len(list(transition_graph.predecessors(dst))) == 1:
                # calls in the middle of blocks OR boring jumps
                # XXX: patch is here
                _merge_ail_nodes(transition_graph, src, dst)
                break

            # calls to functions with no return
            elif type_ == "call":
                transition_graph.remove_node(dst)
                break
        else:
            break

    return transition_graph



import networkx as nx
from cfgutils.os_utils import timeout

#
# Isomorphism
#

def fast_is_isomorphic(g1, g2):
    if len(g1.nodes) != len(g2.nodes):
        return False
    if len(g1.edges) != len(g2.edges):
        return False

    is_iso = False
    try:
        with timeout(seconds=2):
            is_iso = nx.is_isomorphic(g1, g2)
    except TimeoutError:
        pass

    return is_iso



from cfgutils.similarity.ged.abu_aisheh_ged import (
    ged_exact, ged_max, ged_upperbound, ged_explained, graph_edit_distance_core_analysis
)
from cfgutils.similarity.ged.basque_cfged import cfg_edit_distance
from cfgutils.similarity.ged.hu_cfged import hu_cfged
from cfgutils.similarity.block_matcher_base import BlockMatcherBase



mask64bit = 0xffffffffffffffff
mask32bit = 0xffffffff

def rotl64(data, n):
    _data = data & mask64bit
    return ((_data & mask64bit) << n) & mask64bit

# Some primes between 2^63 and 2^64 from CityHash.
seed0_ = 0xc3a5c85c97cb3127
seed1_ = 0xb492b66fbe98f273
seed2_ = 0x9ae16a3b2f90404f
k0 = seed0_
k1 = seed1_
k2 = seed2_

from .sim_hasher import FunctionSimHasher



from collections import defaultdict
import logging
import re
from typing import Optional, List, Tuple

import networkx as nx
import itertools

from cfgutils.sorting import quasi_topological_sort_nodes
from . import seed0_, seed1_, seed2_, rotl64, mask64bit
from ...data import GenericBlock

_l = logging.getLogger(__name__)

CONST_RE = r"(?:\W|0x|^)([0-9a-fA-F]+)(?:h|\W|$)"


class FlowGraph:
    """
    A helper class that wraps networkx graphs to make the following attributes easier to access:
    - subgraph iteration
    - op iterations
    - immediate iteration
    """
    def __init__(self, graph: nx.DiGraph):
        self.graph = graph

        self._top_ordered_nodes = quasi_topological_sort_nodes(graph)
        self._addr_ordered_nodes = sorted(list(self.graph.nodes), key=lambda x: x.addr)
        self.statements = list(itertools.chain.from_iterable([blk.statements for blk in self._addr_ordered_nodes]))
        self._in_degrees = {node: graph.in_degree(node) for node in self._top_ordered_nodes}
        self._out_degrees = {node: graph.out_degree(node) for node in self._top_ordered_nodes}

        self.nodes_and_distance: List[Tuple[GenericBlock, int]] = self._build_nodes_and_distances()
        self.mnemonic_ngrams: List[Tuple[str, str, str]] = self.BuildMnemonicNgrams()
        self.immediates: List[int] = self.FindImmediateValues()

    def _build_nodes_and_distances(self):
        ones, twos, threes = [], [], []
        for node in self.graph.nodes:
            ones.append((node, 1))
            twos.append((node, 2))
            threes.append((node, 3))

        return ones + twos + threes

    def BuildMnemonicNgrams(self):
        mnemonics = [stmt.op for stmt in self.statements]
        mnem_tuples = []
        for index, mnem in enumerate(mnemonics):
            if index + 2 >= len(mnemonics):
                break

            mnem_tuples.append((
                mnem,
                mnemonics[index + 1],
                mnemonics[index + 2],
            ))

        return mnem_tuples

    def FindImmediateValues(self):
        """
        Only consider immediates as useful that are either greater than
        0x4000 or (not divisible by 4 and greater 10). This should remove
        most stack offsets.

        These are precisely the heuristics that should be removed by the
        machine-learning step, but since the baseline is supposed to work
        reasonably well even without the learning step, we need such stuff
        here.

        Also removes data structure offsets, though.
        """
        immediates = []
        for operand in itertools.chain.from_iterable([stmt.operands for stmt in self.statements]):
            _imms = self.ExtractImmediateFromString(operand)
            if not _imms:
                continue

            for _imm in _imms:
                # TODO: just a rule from the original implementation... maybe remove it to get smaller imms
                if (abs(_imm) > 0x4000) or ((_imm % 4 != 0) and (_imm > 10)):
                    immediates.append(_imm)

        return immediates

    def ExtractImmediateFromString(self, string):
        """
        Extracts all the Hex Digits from a string.
        """
        if isinstance(string, int):
            return [string]
        if not isinstance(string, str):
            return []

        imms = []
        for imm in re.findall(CONST_RE, string):
            try:
                val = int(imm, 16)
            except ValueError:
                continue

            imms.append(val)
        return imms

    def CalculateHash(self, start_node=None, k0=0xc3a5c85c97cb3127, k1=0xb492b66fbe98f273, k2=0x9ae16a3b2f90404f):
        """
        TODO: test this against the original function output!
        """
        if start_node is None:
            _starts = list(node for node in self.graph.nodes if self.graph.in_degree(node) == 0)
            if len(_starts) != 1:
                raise ValueError("Graph must have exactly one start node is none is provided")

            start_node = _starts[0]

        if start_node not in self.graph.nodes:
            raise ValueError("Start node must be in the graph")

        # compute the topological order of the graph and reconstruct that order
        # for both the forward and backward edges based on the original code.
        out_edges = defaultdict(list)
        ordered_nodes = self._top_ordered_nodes[self._top_ordered_nodes.index(start_node):]
        order_forward = {ordered_nodes[0]: 0}  # computed from out edges
        order_backward = {}  # computed from in edges
        order_both = {}  # computed from both edges
        back_idx = 0
        fwd_idx = 1
        bi_idx = 0
        for node in ordered_nodes:
            for pred in self.graph.predecessors(node):
                if pred not in order_backward:
                    order_backward[pred] = back_idx
                    back_idx += 1
                if pred not in order_both:
                    order_both[pred] = bi_idx
                    bi_idx += 1
                    order_both[node] = bi_idx
                    bi_idx += 1
            for succ in self.graph.successors(node):
                out_edges[node].append(succ)
                if succ not in order_forward:
                    order_forward[succ] = fwd_idx
                    fwd_idx += 1
                if succ not in order_both:
                    order_both[succ] = bi_idx
                    bi_idx += 1
                    order_both[node] = bi_idx
                    bi_idx += 1

        for node in ordered_nodes:
            if node not in order_backward:
                order_backward[node] = back_idx
                back_idx += 1

        hash_result = 0x0BADDEED600DDEED
        for source, dst_nodes in out_edges.items():
            per_edge_hash = 0x600DDEED0BADDEED
            for target in dst_nodes:
                per_edge_hash += (k0 * order_forward[source]) & mask64bit
                per_edge_hash = rotl64(per_edge_hash, 7)
                per_edge_hash += (k1 * order_backward[source]) & mask64bit
                per_edge_hash = rotl64(per_edge_hash, 7)
                per_edge_hash += (k2 * order_both[source]) & mask64bit
                per_edge_hash = rotl64(per_edge_hash, 7)
                per_edge_hash += (k0 * self._in_degrees[source]) & mask64bit
                per_edge_hash = rotl64(per_edge_hash, 7)
                per_edge_hash += (k1 * self._out_degrees[source]) & mask64bit
                per_edge_hash = rotl64(per_edge_hash, 7)

                per_edge_hash += (k2 * order_forward[target]) & mask64bit
                per_edge_hash = rotl64(per_edge_hash, 7)
                per_edge_hash += (k0 * order_backward[target]) & mask64bit
                per_edge_hash = rotl64(per_edge_hash, 7)
                per_edge_hash += (k1 * order_both[target]) & mask64bit
                per_edge_hash = rotl64(per_edge_hash, 7)
                per_edge_hash += (k2 * self._in_degrees[target]) & mask64bit
                per_edge_hash = rotl64(per_edge_hash, 7)
                per_edge_hash += (k0 * self._out_degrees[target]) & mask64bit
                per_edge_hash = rotl64(per_edge_hash, 7)

            hash_result += per_edge_hash

        return hash_result

    @staticmethod
    def GetSubgraph(graph: nx.DiGraph, node: GenericBlock, distance, max_size=30) -> Optional["FlowGraph"]:
        assert node in graph

        total_size = len(node.statements)
        new_nodes = [node]
        # TODO: add sorter for neighbors
        for src, succs in nx.bfs_successors(graph, node, depth_limit=distance):
            succ: GenericBlock
            for succ in succs:
                new_nodes.append(succ)
                total_size += len(succ.statements)

                if total_size > max_size:
                    _l.debug(f"Max size hit for the graph starting with %s of depth %s", node, distance)
                    return None

        return FlowGraph(nx.subgraph(graph, new_nodes))



# This is a very hacky way to convert AIL statements into the pretty version you normally see inside of
# angr-management. This script was extracted from the source of angr-management and modified to work
# outside of the angr-management environment. This is not a good way to do this, but it works for now.
#

import logging
from typing import Any

import ailment

l = logging.getLogger(__name__)


#
# Utility
#

def is_printable(ch):
    return 32 <= ch < 127


def filter_string_for_display(s):
    output = ""
    for ch in s.replace("\r", "\\r").replace("\n", "\\n").replace("\t", "\\t"):
        char = ord(ch)
        if not is_printable(char):
            ch = "\\x%0.2x" % char
        output += ch
    return output


def fast_memory_load_pointer(project, addr, size=None):
    try:
        return project.loader.memory.unpack_word(addr, size=size)
    except KeyError:
        return None


def string_at_addr(cfg, addr, project, max_size=50):
    try:
        mem_data = cfg.memory_data[addr]
    except KeyError:
        return None

    if mem_data.sort == "string":
        str_content = mem_data.content.decode("utf-8")
    elif mem_data.sort == "pointer-array":
        ptr = fast_memory_load_pointer(project, mem_data.address)
        try:
            next_level = cfg.memory_data[ptr]
        except KeyError:
            return None

        if next_level.sort != "string":
            return None

        str_content = next_level.content.decode("utf-8")
    else:
        return None

    if str_content is not None:
        if len(str_content) > max_size:
            return '"' + filter_string_for_display(str_content[:max_size]) + '..."'
        else:
            return '"' + filter_string_for_display(str_content) + '"'
    else:
        return None


#
# Pretty Printing Classes
#

class OutputBuffer:
    def __init__(self):
        self.text = ""
        self.position = 0

    def insertText(self, text):
        self.text += text
        self.position += len(text)

    def newline(self):
        self.insertText("\n")


class PrettyBlockCodeObj:
    def __init__(self, obj: Any, project: Any, cfg: Any, *args, **kwargs):
        self.obj = obj
        self.project = project
        self.cfg = cfg
        self.span = None
        self.subobjs = []
        self.create_subobjs(obj)

    def create_subobjs(self, obj: Any):
        raise NotImplementedError()

    def recreate_subobjs(self):
        self.subobjs.clear()
        self.create_subobjs(self.obj)

    def render(self, output_buffer):
        """
        Add each subobject to the document
        """
        self.recreate_subobjs()
        span_min = output_buffer.position
        for obj in self.subobjs:
            if type(obj) is str:
                output_buffer.insertText(obj)
            else:
                obj.render(output_buffer)
        span_max = output_buffer.position
        self.span = (span_min, span_max)

    def _add_subobj(self, subobj):
        self.subobjs.append(subobj)

    def add_text(self, text: str):
        """
        Add a text leaf
        """
        self._add_subobj(text)

    def add_variable(self, var):
        self._add_subobj(PrettyVariableObj(var, self.project, self.cfg))


class PrettyAilObj(PrettyBlockCodeObj):
    """
    Renders an AIL object
    """

    def __init__(self, obj: Any, project: Any, cfg: Any, *args, stmt=None, **kwargs):
        self.stmt = stmt or obj
        super().__init__(obj, project, cfg, *args, **kwargs)

    def create_subobjs(self, obj: Any):
        self.add_ailobj(obj)

    def add_ailobj(self, obj: Any):
        """
        Map appropriate AIL type to the display type
        """
        subobjcls = {
            ailment.statement.Assignment: PrettyAilAssignmentObj,
            ailment.statement.Store: PrettyAilStoreObj,
            ailment.statement.Jump: PrettyAilJumpObj,
            ailment.statement.ConditionalJump: PrettyAilConditionalJumpObj,
            ailment.statement.Return: PrettyAilReturnObj,
            ailment.statement.Call: PrettyAilCallObj,
            ailment.expression.Const: PrettyAilConstObj,
            ailment.expression.Tmp: PrettyAilTmpObj,
            ailment.expression.Register: PrettyAilRegisterObj,
            ailment.expression.UnaryOp: PrettyAilUnaryOpObj,
            ailment.expression.BinaryOp: PrettyAilBinaryOpObj,
            ailment.expression.Convert: PrettyAilConvertObj,
            ailment.expression.Load: PrettyAilLoadObj,
        }.get(type(obj), PrettyAilTextObj)
        subobj = subobjcls(obj, self.project, self.cfg, stmt=self.stmt)
        self._add_subobj(subobj)


class PrettyAilTextObj(PrettyAilObj):
    """
    Renders an AIL object via __str__
    """

    def create_subobjs(self, obj: Any):
        self.add_text(str(obj))


class PrettyAilAssignmentObj(PrettyAilTextObj):
    """
    Renders an ailment.statement.Assignment
    """

    def create_subobjs(self, obj: ailment.statement.Assignment):
        self.add_ailobj(obj.dst)
        self.add_text(" = ")
        self.add_ailobj(obj.src)


class PrettyAilStoreObj(PrettyAilTextObj):
    """
    Renders an ailment.statement.Store
    """

    def create_subobjs(self, obj: ailment.statement.Store):
        #if obj.variable is None or not self.options.show_variables:
        if obj.variable is None:
            self.add_text("*(")
            self.add_ailobj(obj.addr)
            self.add_text(") = ")
            self.add_ailobj(obj.data)
        else:
            self.add_variable(obj.variable)
            self.add_text(" = ")
            self.add_ailobj(obj.data)


class PrettyAilJumpObj(PrettyAilTextObj):
    """
    Renders an ailment.statement.Jump
    """

    def create_subobjs(self, obj: ailment.statement.Jump):
        self.add_text("goto ")
        self.add_ailobj(obj.target)


class PrettyAilConditionalJumpObj(PrettyAilTextObj):
    """
    Renders an ailment.statement.ConditionalJump
    """

    def create_subobjs(self, obj: ailment.statement.ConditionalJump):
        self.add_text("if ")
        self.add_ailobj(obj.condition)

        #if self.options.show_conditional_jump_targets:
        self.add_text(" goto ")
        self.add_ailobj(obj.true_target)
        self.add_text(" else goto ")
        self.add_ailobj(obj.false_target)


class PrettyAilReturnObj(PrettyAilTextObj):
    """
    Renders an ailment.statement.Return
    """

    def create_subobjs(self, obj: ailment.statement.Return):
        self.add_text("return ")
        for expr in obj.ret_exprs:
            self.add_ailobj(expr)


class PrettyAilCallObj(PrettyAilTextObj):
    """
    Renders an ailment.statement.Call
    """

    def create_subobjs(self, obj: ailment.statement.Call):
        if obj.ret_expr is not None and self.stmt is self.obj:
            self.add_ailobj(obj.ret_expr)
            self.add_text(" = ")
        self.add_ailobj(obj.target)
        self.add_text("(")
        if obj.args:
            for i, arg in enumerate(obj.args):
                if i > 0:
                    self.add_text(", ")
                self.add_ailobj(arg)
        self.add_text(")")


class PrettyAilConstObj(PrettyAilTextObj):
    """
    Renders an ailment.expression.Const
    """

    def create_subobjs(self, obj: ailment.expression.Const):
        # take care of labels first
        kb = self.project.kb
        if obj.value in kb.labels:
            self.add_text(kb.labels[obj.value])
            return

        data_str = string_at_addr(
            self.cfg,
            obj.value,
            self.project,
        )
        if data_str:
            self.add_text(data_str)
        else:
            self.add_text(f"{obj.value:#x}")


class PrettyAilTmpObj(PrettyAilTextObj):
    """
    Renders an ailment.expression.Tmp
    """
    pass


class PrettyAilRegisterObj(PrettyAilTextObj):
    """
    Renders an ailment.expression.Register
    """
    def create_subobjs(self, obj: ailment.expression.Register):
        #if obj.variable is not None and self.options.show_variables:
        if obj.variable is not None:
            self.add_variable(obj.variable)
        else:
            s = f"{obj.reg_name}" if hasattr(obj, "reg_name") else "reg_%d<%d>" % (obj.reg_offset, obj.bits // 8)
            self.add_text(s)


class PrettyAilUnaryOpObj(PrettyAilTextObj):
    """
    Renders an ailment.expression.UnaryOp
    """

    def create_subobjs(self, obj: ailment.expression.UnaryOp):
        self.add_text("(")
        self.add_text(obj.op + " ")
        self.add_ailobj(obj.operand)
        self.add_text(")")


class PrettyAilBinaryOpObj(PrettyAilTextObj):
    """
    Renders an ailment.expression.BinaryOp
    """

    def create_subobjs(self, obj: ailment.expression.BinaryOp):
        self.add_text("(")
        self.add_ailobj(obj.operands[0])
        verbose_op = obj.OPSTR_MAP.get(obj.verbose_op, obj.verbose_op)
        if verbose_op is None:
            verbose_op = "unknown_op"
        self.add_text(" " + verbose_op + " ")
        self.add_ailobj(obj.operands[1])
        self.add_text(")")


class PrettyAilConvertObj(PrettyAilTextObj):
    """
    Renders an ailment.expression.Convert
    """

    def create_subobjs(self, obj: ailment.expression.Convert):
        self.add_text("Conv(%d->%d, " % (obj.from_bits, obj.to_bits))
        self.add_ailobj(obj.operand)
        self.add_text(")")


class PrettyAilLoadObj(PrettyAilTextObj):
    """
    Renders an ailment.expression.Load
    """

    def create_subobjs(self, obj: ailment.expression.Load):
        #if obj.variable is not None and self.options.show_variables:
        if obj.variable is not None:
            self.add_variable(obj.variable)
        else:
            self.add_text("*(")
            self.add_ailobj(obj.addr)
            self.add_text(")")


class PrettyVariableObj(PrettyBlockCodeObj):
    """
    Renders a variable
    """
    def create_subobjs(self, obj):
        #ident = "<%s>" % (obj.ident if obj.ident else "") if self.options.show_variable_identifiers else ""
        ident = ""
        self.add_text(obj.name + ident)


def stmt_to_pretty_text(stmt, project, proj_cfg):
    output_buffer = OutputBuffer()
    pretty_obj = PrettyAilObj(stmt, project, proj_cfg)
    pretty_obj.render(output_buffer)
    return output_buffer.text



import sys
import unittest

from cfgutils.data import numbered_edges_to_block_graph
from cfgutils.similarity.ged.abu_aisheh_ged import ged_exact, ged_max, ged_explained
from cfgutils.similarity.ged.basque_cfged import cfg_edit_distance
from cfgutils.similarity.ged.hu_cfged import hu_cfged


class TestGraphEditDistance(unittest.TestCase):
    def test_explained_ged(self):
        """
        This tests utilizes the graph shown in `test_exact_ged_cfg` and validates the explanation function
        to match the expected edits as shown in the function docstring.
        """

        g1, g2 = CROSS_JUMP_OPT_GRAPHS
        human_readable_edits, edits = ged_explained(g1, g2, print_explanation=True, only_addrs=True)
        real_ged_score = ged_exact(g1, g2)

        # gather expected edits from docstring
        tmp = self.test_exact_ged_cfg.__doc__.split("EDITS:\n")[1].split("\n")[0].strip().split("), ")
        expected_edits = [t + ")" if not t.endswith(")") else t for t in tmp]

        assert len(human_readable_edits) == real_ged_score
        assert all([edit in expected_edits for edit in human_readable_edits])

    def test_exact_ged_cfg(self):
        """
        This function validates that our GED algorithm is working as expected for CFGs. It's important to note
        that our GED algorithm is different from the traditional GED algorithm. We special rules to make it work
        with CFGs. For example, you are not allowed to ever substitute the start or end nodes of a CFG.

        The graphs bellow is based on the Cross Jump Optimization shown in the SAILR paper. g1 is the original CFG
        as you would see in a real program. g2 is the optimized version of g1, usually produced by `-O2` or `-O3`.

        g1:

               1
             /   \
            2     3
            |     | \
            |     |   5
            |     |  /
           6.1    6.2

        g2:

               1
             /   \
            2     3
            |    / \
            |    4  5
            |    \  /
            + ---> 6

        Since changing g2 to g1 is simpler, we will use that as the example. The expected edit distance is 5. The
        expected edits are as follows.
        EDITS:
        del(3, 6), del(6), ins(4), ins(3, 4), ins(4, 6)

        Those edits can be produced by using the `ged_explained` function.
        """
        g1, g2 = CROSS_JUMP_OPT_GRAPHS

        one_to_two_edit_distance = ged_exact(g1, g2, with_timeout=20)
        two_to_one_edit_distance = ged_exact(g2, g1, with_timeout=20)

        assert one_to_two_edit_distance == 5
        # note: rooted graph edit distance algorithms are not the same computed both ways
        assert two_to_one_edit_distance == 7


class TestBasqueCFGED(unittest.TestCase):
    def test_exact_cfged(self):
        """
        The CFGED algorithm is a special version of GED that is designed to work with CFGs, quickly. To do this,
        we use node label information to point the algorithm to what we believe is matching regions. This allows us
        to run GED on the small regions, then sum up all the region scores. In the best case, the CFGED score is the
        same as the exact GED score. In the worst case, it's the max GED score. At all times, the score should be
        between the max and exact scores.

        For this test, we use the same graphs from `TestGraphEditDistance.test_exact_ged_cfg`.
        Turning g1 -> g2, we expect the CFGED score to be the exact score, 5. The other way around, we expect it to
        maintain its earlier stated bounds.
        """
        g1, g2 = CROSS_JUMP_OPT_GRAPHS

        # This map is critical to using the algorithm. You don't need to provide a mapping for every node (although
        # you can, and it won't hurt). You only need to provide mappings for tail regions, like node 3 in both graphs.
        # Since we know how every node maps to each other, we provide a map for all of them.
        node_map = {x: {x} for x in range(7)}

        combos = [(g1, g2), (g2, g1)]
        for i, graphs in enumerate(combos):
            first_g, second_g = graphs
            max_ged_score = ged_max(first_g, second_g)
            exact_ged_score = ged_exact(first_g, second_g)
            # the node map is the same for both graphs, so we can use it for both
            cfged_score = cfg_edit_distance(first_g, second_g, node_map, node_map)
            assert exact_ged_score <= cfged_score <= max_ged_score

            if i == 0:
                assert exact_ged_score == cfged_score


class TestHuCFGED(unittest.TestCase):
    def test_cross_jump_graphs(self):
        g1, g2 = CROSS_JUMP_OPT_GRAPHS
        score = hu_cfged(g1, g2)
        real_score = ged_exact(g1, g2)
        max_score = ged_max(g1, g2)
        print(f"score={score}, real_score={real_score}, max_score={max_score}")
        assert real_score <= score <= max_score
#
# Some common graphs used among testcases
#


CROSS_JUMP_OPT_GRAPHS = (
    # g1
    numbered_edges_to_block_graph(
        [(1, 2), (1, 3), (3, 6.2), (3, 5), (5, 6.2), (2, 6.1)]
    ),
    # g2
    numbered_edges_to_block_graph(
        [(1, 2), (1, 3), (3, 4), (3, 5), (4, 6), (5, 6), (2, 6)]
    ),
)


if __name__ == "__main__":
    unittest.main(argv=sys.argv)
